"""Server.sc"""

import enum
import subprocess
import threading
import logging
import os
import pathlib
import time
import errno

from ..base import main as _libsc3
from ..base import utils as utl
from ..base import classlibrary as clb
from ..base import netaddr as nad
from ..base import model as mdl
from ..base import systemactions as sac
from ..base import functions as fn
from ..base import platform as plf
from ..base import stream as stm
from ..base import clock as clk
from . import _engine as eng
from . import synthdef as sdf
from . import _serverstatus as sst
from . import _nodewatcher as ndw
from . import _volume as vlm
from . import recorder as rcd
from . import node as nod
from . import bus
from . import _graphparam as gpp


__all__ = ['s', 'Server', 'ServerOptions']


_logger = logging.getLogger(__name__)


s = None
'''Default server global variable.'''


class Defaults(enum.Enum):
    UDP_PORT = ('-u', None)  # mutex -t, opt int
    TCP_PORT = ('-t', None)  # mutex -u, opt int

    BIND_ADDRESS = ('-B', '127.0.0.1')  # opt str
    MAX_LOGINS = ('-l', 64)  # opt int
    PASSWORD = ('-p', None)  # opt str
    ZEROCONF = ('-R', 1)  # RENDEZVOUS, boolean

    RESTRICTED_PATH = ('-P', None)  # opt str
    UGEN_PLUGINS_PATH = ('-U', None)  # opt str

    CONTROL_BUSES = ('-c', 16384)  # # opt int, num_control_bus_channels
    AUDIO_BUSES = ('-a', 1024)  # # opt int, num_audio_bus_channels
    INPUT_CHANNELS = ('-i', 8)  # opt int, num_input_bus_channels
    OUTPUT_CHANNELS = ('-o', 8)  # opt int, num_output_bus_channels
    BLOCK_SIZE = ('-z', 64)  # opt int
    BUFFERS = ('-b', 1024)  # opt int, num_buffers
    MAX_NODES = ('-n', 1024)  # opt int
    MAX_SYNTHDEFS = ('-d', 1024)  # opt int, max_synth_defs
    RT_MEMORY = ('-m', 8192)  # opt int, mem_size
    WIRES = ('-w', 64)  # opt int, num_wire_bufs
    RGENS = ('-r', 64)  # opt int, num_rgens
    LOAD_SYNTHDEFS = ('-D', 1)  # opt bool, load_defs

    HW_DEVICE_NAME = ('-H', None)  # opt str
    HW_BUFFER_SIZE = ('-Z', 0)  # opt int, hardware_buffer_size
    SAMPLE_RATE = ('-S', 0)  # opt int, supernova differs: 44100
    NRT = ('-N', None)  # mode flag
    VERBOSE = ('-V', 0)  # opt int

    # Supernova only.
    MEMORY_LOCKING = ('-L', None)  # mode flag
    THREADS = ('-T', 4)  # opt int
    USE_SYSTEM_CLOCK = ('-C', 0)  # opt bool

    # Darwin only.
    INPUT_STREAMS = ('-I', None)  # opt str, input_streams_enabled
    OUTPUT_STREAMS = ('-O', None)  # opt str, output_streams_enabled

    def __init__(self, flag, default):
        self.flag = flag
        self.default = default


class ServerOptions():
    # locals().update(Defaults.__members__)

    def __init__(self):
        self.program = plf.Platform.default_server_cmd
        self.protocol = 'upd'

        self.bind_address = Defaults.BIND_ADDRESS.default
        self.max_logins = 1  # not default cmd value.
        self.password = Defaults.PASSWORD.default
        self.zeroconf = 0  # not default cmd value.

        self.restricted_path = Defaults.RESTRICTED_PATH.default
        self.ugen_plugins_path = Defaults.UGEN_PLUGINS_PATH.default

        self.control_buses = Defaults.CONTROL_BUSES.default
        self.audio_buses = Defaults.AUDIO_BUSES.default
        self.input_channels = 2  # not default cmd value
        self.output_channels = 2  # not default cmd value
        self.block_size = Defaults.BLOCK_SIZE.default
        self.buffers = Defaults.BUFFERS.default
        self.max_nodes = Defaults.MAX_NODES.default
        self.max_synthdefs = Defaults.MAX_SYNTHDEFS.default
        self.rt_memory = Defaults.RT_MEMORY.default
        self.wires = Defaults.WIRES.default
        self.rgens = Defaults.RGENS.default
        self.load_synthdefs = Defaults.LOAD_SYNTHDEFS.default

        # removed sclang's inDevice, outDevice, prListDevices, etc., the
        # solution would be to use an IOAudioDevice class that generate
        # the string for this options.
        self.hw_device_name = Defaults.HW_DEVICE_NAME.default
        self.hw_buffer_size = Defaults.HW_BUFFER_SIZE.default
        self.sample_rate = Defaults.SAMPLE_RATE.default
        # self.nrt = Defaults.NRT.default  # See options_list
        self.verbose = Defaults.VERBOSE.default

        # Supernova only.
        self.memory_locking = Defaults.MEMORY_LOCKING.default
        self.threads = Defaults.THREADS.default
        self.use_system_clock = Defaults.USE_SYSTEM_CLOCK.default

        # Darwin only.
        self.input_streams = Defaults.INPUT_STREAMS.default
        self.output_streams = Defaults.OUTPUT_STREAMS.default

        # Language side.
        self.reserved_audio_buses = 0  # not used, really.
        self.reserved_control_buses = 0  # not used, really.
        self.reserved_buffers = 0  # not used, really.
        self.initial_node_id = 1000
        # self.remote_control_volume = False  # ServerPlusGui makeGui, no GUI.
        self.rec_header_format = 'aiff'
        self.rec_sample_format = 'float'
        self.rec_channels = 2
        self.rec_buf_size = None

    def options_list(self, port, osc_file=None, input_file=None,
                     output_file=None):
        o = []

        if port is None:
            o.append(Defaults.NRT.flag)
            if osc_file is None:
                raise ValueException('osc_file must be supplied for nrt')
            o.append(str(osc_file))
            input_file = '_' if input_file is None else str(input_file)
            o.append(input_file)
            if output_file is None:
                # TODO: Method in Platform for name generation. See Recorder & Buffer.
                output_file = plf.Platform.recording_dir
                output_file.mkdir(exist_ok=True)
                output_file = str(output_file / 'SC_')
                output_file += time.strftime('%Y%m%d_%H%M%S')
                output_file += '.' + self.rec_header_format
            else:
                output_file = str(output_file)
            o.append(output_file)
            # No default SR for NRT.
            o.append(str(self.sample_rate) if self.sample_rate else '44100')
            o.append(self.rec_header_format)
            o.append(self.rec_sample_format)
        else:
            if self.protocol == 'tcp':
                o.extend([Defaults.TCP_PORT.flag, str(port)])
            else:
                o.extend([Defaults.UDP_PORT.flag, str(port)])

            if self.bind_address != Defaults.BIND_ADDRESS.default:
                o.extend([Defaults.BIND_ADDRESS.flag, str(self.bind_address)])
            if self.max_logins != Defaults.MAX_LOGINS.default:
                o.extend([Defaults.MAX_LOGINS.flag, str(self.max_logins)])
            if self.password != Defaults.PASSWORD.default:
                o.extend([Defaults.PASSWORD.flag, str(self.password)])
            if self.zeroconf != Defaults.ZEROCONF.default:
                o.extend([Defaults.ZEROCONF.flag, str(int(self.zeroconf))])

            if self.sample_rate != Defaults.SAMPLE_RATE.default:
                o.extend([Defaults.SAMPLE_RATE.flag, str(self.sample_rate)])

        if self.restricted_path != Defaults.RESTRICTED_PATH.default:
            o.extend([Defaults.RESTRICTED_PATH.flag, str(self.restricted_path)])
        if self.ugen_plugins_path != Defaults.UGEN_PLUGINS_PATH.default:
            plist = utl.as_list(self.ugen_plugins_path)
            plist = [os.path.normpath(x) for x in plist]
            o.extend([Defaults.UGEN_PLUGINS_PATH.flag, ':'.join(plist)])

        if self.control_buses != Defaults.CONTROL_BUSES.default:
            o.extend([Defaults.CONTROL_BUSES.flag, str(self.control_buses)])
        if self.audio_buses != Defaults.AUDIO_BUSES.default:
            o.extend([Defaults.AUDIO_BUSES.flag, str(self.audio_buses)])
        if self.input_channels != Defaults.INPUT_CHANNELS.default:
            o.extend([Defaults.INPUT_CHANNELS.flag, str(self.input_channels)])
        if self.output_channels != Defaults.OUTPUT_CHANNELS.default:
            o.extend([Defaults.OUTPUT_CHANNELS.flag, str(self.output_channels)])

        if self.block_size != Defaults.BLOCK_SIZE.default:
            o.extend([Defaults.BLOCK_SIZE.flag, str(self.block_size)])
        if self.buffers != Defaults.BUFFERS.default:
            o.extend([Defaults.BUFFERS.flag, str(self.buffers)])
        if self.max_nodes != Defaults.MAX_NODES.default:
            o.extend([Defaults.MAX_NODES.flag, str(self.max_nodes)])
        if self.max_synthdefs != Defaults.MAX_SYNTHDEFS.default:
            o.extend([Defaults.MAX_SYNTHDEFS.flag, str(self.max_synthdefs)])
        if self.rt_memory != Defaults.RT_MEMORY.default:
            o.extend([Defaults.RT_MEMORY.flag, str(self.rt_memory)])
        if self.wires != Defaults.WIRES.default:
            o.extend([Defaults.WIRES.flag, str(self.wires)])
        if self.rgens != Defaults.RGENS.default:
            o.extend([Defaults.RGENS.flag, str(self.rgens)])
        if self.load_synthdefs != Defaults.LOAD_SYNTHDEFS.default:
            o.extend([Defaults.LOAD_SYNTHDEFS.flag,
                str(int(self.load_synthdefs))])

        if self.hw_device_name != Defaults.HW_DEVICE_NAME.default:
            o.extend([Defaults.HW_DEVICE_NAME.flag, str(self.hw_device_name)])
        if self.hw_buffer_size != Defaults.HW_BUFFER_SIZE.default:
            o.extend([Defaults.HW_BUFFER_SIZE.flag, str(self.hw_buffer_size)])

        if self.verbose != Defaults.VERBOSE.default:
            o.extend([Defaults.VERBOSE.flag, str(self.verbose)])

        # Supernova only.
        if self.program == plf.Platform.SUPERNOVA_CMD:
            if self.memory_locking != Defaults.MEMORY_LOCKING.default:
                o.append(Defaults.MEMORY_LOCKING.flag)
            if self.threads != Defaults.THREADS.default:
                o.extend([Defaults.THREADS.flag, str(self.threads)])
            if self.use_system_clock != Defaults.USE_SYSTEM_CLOCK.default:
                o.extend([Defaults.USE_SYSTEM_CLOCK.flag,
                    str(int(self.use_system_clock))])

        # Darwin only.
        if plf.Platform.name.startswith('darwin'):
            if self.input_streams != Defaults.INPUT_STREAMS.default:
                o.extend([Defaults.INPUT_STREAMS.flag, str(self.input_streams)])
            if self.output_streams != Defaults.OUTPUT_STREAMS.default:
                o.extend([Defaults.OUTPUT_STREAMS.flag,
                    str(self.output_streams)])

        return o

    def first_private_bus(self):
        return self.output_channels + self.input_channels


class ServerShmInterface():
    def __init__(self, port):
        self.ptr = None # variable de bajo nivel, depende de la implementación.
        self.finalizer = None # variable de bajo nivel, depende de la implementación.
        self.connect(port) # llama a una primitiva y debe guardar port a bajo nivel

    # copy # // never ever copy! will cause duplicate calls to the finalizer!
    def connect(self, port): ... # primitiva
    def disconnect(self): ... # primitiva
    def get_control_bus_value(self): ... # primitiva
    def get_control_bus_values(self): ... # primitiva
    def set_control_bus_value(self, value): ... # primitiva # BUG: desconozco los parámetros.
    def set_control_bus_values(self, *values): ... # primitiva # BUG: desconozco los parámetros.


class ServerProcess():
    def __init__(self, on_exit=None):
        self.on_exit = on_exit or (lambda exit_code: None)
        self.proc = None
        self.timeout = 0.1
        self._detached = False

    def run(self, server):
        if self.proc:
            return

        cmd = [server.options.program]
        cmd.extend(server.options.options_list(server.addr.port))

        self.proc = subprocess.Popen(
            cmd,  # TODO: maybe a method popen_cmd regarding server, options and platform.
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,
            universal_newlines=True)
        self._redirect_outerr()

        def popen_wait_thread():
            self.proc.wait()
            self._tflag.set()
            # self._tout.join()
            # self._terr.join()
            if not self._detached:
                self.on_exit(self.proc.poll())

        t = threading.Thread(
            target=popen_wait_thread,
            name=f'{type(self).__name__}.popen_wait id: {id(self)}')
        t.daemon = True
        t.start()

    def running(self):
        if self.proc is None or self._detached:
            return False
        return self.proc.poll() is None

    def finish(self):
        if self._detached:
            return

        def terminate_proc_thread():
            try:
                if self.running():
                    self.proc.terminate()
                    self.proc.wait(timeout=self.timeout)
            except subprocess.TimeoutExpired:
                self.proc.kill()
                self.proc.communicate() # just to be polite

        t = threading.Thread(
            target=terminate_proc_thread,
            name=f'{type(self).__name__}.terminate id: {id(self)}')
        t.daemon = True
        t.start()

    def detach(self):
        self._tflag.set()
        self._detached = True

    def _redirect_outerr(self):
        def read(out, flag, logger):
            while not flag.is_set() and self.running():  # BUG: is still a different thread.
                line = out.readline()
                if line:
                    # print(line, end='')
                    logger.info(line.rstrip())

        def make_thread(out, flag, out_name):
            logger = logging.getLogger(f'SERVER.{out_name}')
            t = threading.Thread(
                target=read,
                name=f'{type(self).__name__}.{out_name} id: {id(self)}',
                args=(out, flag, logger))
            t.daemon = True
            t.start()
            return t

        self._tflag = threading.Event()
        self._tout = make_thread(self.proc.stdout, self._tflag, 'stdout')
        self._terr = make_thread(self.proc.stderr, self._tflag, 'stderr')


class MetaServer(type):
    def __init__(cls, *_):
        def init_func(cls):
            cls.DEFAULT_ADDRESS = nad.NetAddr('127.0.0.1', 57110)
            cls.named = dict()
            cls.all = set()
            cls.sync_s = True

            cls._node_alloc_class = eng.NodeIDAllocator
            # // cls._node_alloc_class = ReadableNodeIDAllocator;
            cls._buffer_alloc_class = eng.ContiguousBlockAllocator
            cls._bus_alloc_class = eng.ContiguousBlockAllocator

            cls.default = cls.local = cls('localhost', cls.DEFAULT_ADDRESS)
            # cls.internal = cls(
            #     'internal', nad.NetAddr(None, None))  # No internal by now.

        clb.ClassLibrary.add(cls, init_func)

    @property
    def default(cls):
        return cls._default

    @default.setter
    def default(cls, value):
        cls._default = value
        if cls.sync_s:
            _logger.info(f"setting global variable 's' to '{value.name}'")
            global s
            s = value
        for server in cls.all:
            mdl.NotificationCenter.notify(server, 'default', value)


class Server(gpp.NodeParameter, metaclass=MetaServer):
    def __init__(self, name, addr, options=None):
        super(gpp.NodeParameter, self).__init__(self)

        self.addr = addr  # @property setter
        self._set_name(name)  # Raises ValueException if duplicated.
        type(self).all.add(self)

        # These attributes are initialized through addr setter.
        # self._is_local
        # self._in_process

        self.options = options or ServerOptions()
        self.latency = 0.2
        self.dump_mode = 0

        # These attributes are initialized through self.client_id property setter.
        # self._node_allocator = None # init in _new_node_allocators()
        # self._default_group = None # init in _new_node_allocators() -> _make_default_groups()
        # self._default_groups = None # init in _new_node_allocators() -> _make_default_groups()
        # self._control_bus_allocator = None # init in _new_bus_allocators()
        # self._audio_bus_allocator = None # init in _new_bus_allocators()
        # self._buffer_allocator = None # init in _new_buffer_allocators()
        # self._scope_buffer_allocator = None # init in _new_scope_buffer_allocators()

        self._status_watcher = sst.ServerStatusWatcher(server=self)
        self._node_watcher = ndw.NodeWatcher(server=self)

        self._set_client_id(0)  # Assumed id to work without booting.

        self._volume = vlm.Volume(server=self, persist=True)
        self._recorder = rcd.Recorder(server=self)
        self._recorder.notify_server = True

        self.tree = lambda *args: None

        self._pid = None
        self._shm_interface = None  # ServerShmInterface
        self._server_process = None  # ServerProcess
        self._pid_release_condition = stm.Condition(lambda: self._pid is None)
        mdl.NotificationCenter.notify(type(self), 'server_added', self)

    def remove(self):
        type(self).all.remove(self)
        del type(self).named[self.name]

    @property
    def addr(self):
        return self._addr

    @addr.setter
    def addr(self, value):
        if any(s.addr == value for s in type(self).all):
            raise ValueError(f'{value} already in use by other server')
        self._addr = value
        self._in_process = self._addr.addr == 0
        self._is_local = self._in_process or self._addr.is_local

    # @property
    # def in_process(self):
    #     return self._in_process

    @property
    def is_local(self):
        "Return true if the server is in localhost."
        return self._is_local

    @property
    def name(self):
        return self._name

    def _set_name(self, value):
        # Name can be set only at creation time.
        if value in type(self).named:
            raise ValueError(f"server name '{value}' already exists")
        self._name = value
        type(self).named[value] = self

    @property
    def default_group(self):
        return self._default_group

    @property
    def default_groups(self):
        return self._default_groups

    @property
    def status(self):
        '''
        ServerStatusWatcher instance that keeps track of server status.
        '''
        # This this read-only property (non-data descritor) is the only
        # intended user interface to ServerStatusWatcher instances. Library's
        # style always uses _status_watcher private attribute directly.
        # NodeWatcher, on the other hand, is an implementations detail. Nodes
        # are registered or not with their own interface. Other refinements
        # can be applied later, the idea is not to cram the Server interface
        # and to always use composition instead, e.g. addr, options, status,
        # volume, recorder, etc.
        return self._status_watcher

    @property
    def volume(self):
        return self._volume

    @property
    def recorder(self):
        return self._recorder

    @property
    def pid(self):
        return self._pid


    ### Client ID  ##

    @property
    def client_id(self):
        return self._client_id

    def _set_client_id(self, value):
        if not isinstance(value, int):
            raise TypeError(f'value is not an int: {type(value)}')
        if value < 0 or value >= self.options.max_logins:
            # Supernova ignores max_logins option
            # and doesn't return max_logins info.
            _logger.error(
                f'id ({value}) outside options.max_logins '
                f'({s.options.max_logins}), current id is {self._client_id}')
            if self.options.program == plf.Platform.SUPERNOVA_CMD:
                _logger.info(
                    'NOTE: suepernova servers ignore options.max_logins')
            return
        self._client_id = value
        self._new_allocators()


    ### ClientID-based id allocators ###

    def _new_allocators(self):
        self._new_node_allocators()
        self._new_bus_allocators()
        self._new_buffer_allocators()
        # self._new_scope_buffer_allocators()
        mdl.NotificationCenter.notify(self, '_new_allocators')

    def _new_node_allocators(self):
        self._node_allocator = type(self)._node_alloc_class(
            self.client_id, self.options.initial_node_id)
            #, self._status_watcher.max_logins) # *** BUG: en sclang, _node_alloc_class es eng.NodeIDAllocator por defecto, los alocadores originales reciben 2 parámetros, ContiguousBlockAllocator, que se usa para buses y buffers, recibe uno más, cambia la interfaz. Acá se pasa el tercer parámetro y NodeIDAllocator lo ignora (característica de las funciones de sclang), tengo que ver cómo maneja los ids de los nodos por cliente.
        # // defaultGroup and defaultGroups depend on allocator,
        # // so always make them here:
        self._make_default_groups()

    def _make_default_groups(self):
        # // Keep defaultGroups for all clients on this server.
        self._default_groups = [nod.Group.basic_new(
            self, self._node_allocator.num_ids * client_id + 1
        ) for client_id in range(self._status_watcher.max_logins)]
        self._default_group = self._default_groups[self.client_id]

    def _new_bus_allocators(self):
        audio_bus_io_offset = self.options.first_private_bus()
        num_ctrl_per_client = (
            self.options.control_buses // self._status_watcher.max_logins)
        num_audio_per_client = (
            (self.options.audio_buses - audio_bus_io_offset) //
            self._status_watcher.max_logins)
        ctrl_reserved_offset = self.options.reserved_control_buses
        ctrl_bus_client_offset = num_ctrl_per_client * self.client_id
        audio_reserved_offset = self.options.reserved_audio_buses
        audio_bus_client_offset = num_audio_per_client * self.client_id

        self._control_bus_allocator = type(self)._bus_alloc_class(
            num_ctrl_per_client,
            ctrl_reserved_offset,
            ctrl_bus_client_offset)

        self._audio_bus_allocator = type(self)._bus_alloc_class(
            num_audio_per_client,
            audio_reserved_offset,
            audio_bus_client_offset + audio_bus_io_offset)

    def _new_buffer_allocators(self):
        num_buffers_per_client = (
            self.options.buffers // self._status_watcher.max_logins)
        num_reserved_buffers = self.options.reserved_buffers
        buffer_client_offset = num_buffers_per_client * self.client_id

        self._buffer_allocator = type(self)._buffer_alloc_class(
            num_buffers_per_client, num_reserved_buffers, buffer_client_offset)

    # shm for GUI.
    # def _new_scope_buffer_allocators(self):
    #     if self._is_local:
    #         self._scope_buffer_allocator = eng.StackNumberAllocator(0, 127)

    def next_buffer_number(self, n):
        bufnum = self._buffer_allocator.alloc(n)
        if bufnum is None:
            if n > 1:
                raise Exception(
                    f'No block of {n} consecutive '
                    'buffer numbers is available')
            else:
                raise Exception(
                    'No more buffer numbers, free '
                    'some buffers before allocating')
        return bufnum

    def free_all_buffers(self):
        bundle = []
        for block in self._buffer_allocator.blocks():
            for i in range(block.address, block.address + block.size - 1):
                bundle.append(['/b_free', i])
            self._buffer_allocator.free(block.address)
        self.send_bundle(None, *bundle)

    def next_node_id(self):
        return self._node_allocator.alloc()

    def next_perm_node_id(self):
        return self._node_allocator.alloc_perm()

    def free_perm_node_id(self, id):
        return self._node_allocator.free_perm(id)


    ### Network messages ###

    def send_msg(self, *args):
        self.addr.send_msg(*args)

    def send_bundle(self, time, *elements):
        self.addr.send_bundle(time, *elements)

    def send_synthdef(self, name, dir=None):
        # // Load from disk locally, send remote.
        dir = dir or plf.Platform.synthdef_dir
        dir = pathlib.Path(dir)
        full_path = dir / f'{name}.{sdf.SynthDef._SUFFIX}'
        try:
            with open(full_path, 'rb') as file:
                buffer = file.read()
                self.send_msg('/d_recv', buffer)
        except FileNotFoundError:
            _logger.warning(f'send_synthdef FileNotFoundError: {full_path}')

    def load_synthdef(self, name, completion_msg=None, dir=None):
        # // Tell server to load from disk.
        dir = dir or plf.Platform.synthdef_dir
        dir = pathlib.Path(dir)
        path = str(dir / f'{name}.{sdf.SynthDef._SUFFIX}')
        self.send_msg('/d_load', path, fn.value(completion_msg, self))

    def load_directory(self, dir, completion_msg=None):
        self.send_msg('/d_loadDir', dir, fn.value(completion_msg, self))

    def send_status_msg(self):
        self.addr.send_status_msg()

    def dump_osc(self, code=1):
        # 0 - turn dumping OFF.
        # 1 - print the parsed contents of the message.
        # 2 - print the contents in hexadecimal.
        # 3 - print both the parsed and hexadecimal representations of the contents.
        self.dump_mode = code
        self.send_msg('/dumpOSC', code)
        mdl.NotificationCenter.notify(self, 'dump_osc', code)

    def reorder(self, node_list, target, add_action='addToHead'):
        target = gpp.node_param(target)._as_target()
        node_list = [x.node_id for x in node_list]
        self.send(
            '/n_order', nod.Node.action_number_for(add_action), # 62
            target.node_id, *node_list)

    def sync(self, condition=None, latency=None, elements=None):
        if _libsc3.main is _libsc3.NrtMain:
            yield 0  # *** NOTE: Depends on Condition implementation.
        else:
            yield from self.addr.sync(condition, latency, elements)


    ### Network message bundling ###

    def bind(self):
        # with s.bind():
        return nad.BundleNetAddr(self)


    ### Default group ###

    def init_tree(self):
        def init_task():
            self._send_default_groups()
            self.tree(self)
            yield from self.sync()
            sac.ServerTree.run(self)
            yield from self.sync()

        stm.Routine.run(init_task, clk.AppClock)

    def _send_default_groups(self):
        for group in self._default_groups:
            self.send_msg('/g_new', group.node_id, 0, 0)

    def _send_default_groups_for_client_ids(self, client_ids):  # unused
        for i in client_ids:
            group = self._default_groups[i]
            self.send_msg('/g_new', group.node_id, 0, 0)


    # These atributes are just a wrapper of ServerOptions, use s.options.
    # @property rec_header_format
    # @property rec_sample_format
    # @property rec_channels
    # @property rec_buf_size


    ### Server status ###

    @classmethod
    def _resume_status_threads(cls):  # NOTE: for System Actions.
        for server in cls.all:
            server._status_watcher._resume_alive_thread()


    ### Shared memory interface ###

    def _disconnect_shm(self):  # Was _disconnect_shared_memory
        ...

    def _connect_shm(self):  # Was _connect_shared_memory.
        ...

    @property
    def has_shm(self):  # Was has_shm_interface.
        return self._shm_interface is not None


    ### Boot and login ###

    def register(self, on_complete=None, on_failure=None):
        if self._status_watcher._server_registering:
            _logger.info(f"server '{self.name}' already registeringing")
            return

        if not self._status_watcher.server_running:
            self._status_watcher._server_registering = True

            def _on_complete(server):
                self._status_watcher._server_registering = False
                fn.value(on_complete, self)
                _libsc3.main._atexitq.add(
                    _libsc3.main._atexitprio.SERVERS, self._unregister_atexit)

            def _on_failure(server):
                if self.addr.proto == 'tcp':
                    self.addr.disconnect()
                self._status_watcher._server_registering = False
                fn.value(on_failure, self)

            self._status_watcher._add_action(
                'register', _on_complete, _on_failure)

            if self.options.protocol == 'tcp':
                def success():
                    self._status_watcher._start_alive_thread()
                def failure():
                    self._status_watcher._server_registering = False
                self.addr.connect(success, failure)
            else:
                self._status_watcher._start_alive_thread()
        else:
            _logger.info(f"server '{self.name}' already running")

    def unregister(self, on_complete=None, on_failure=None):
        if self._status_watcher._server_unregistering:
            _logger.info(f"server '{self.name}' already unregisteringing")
            return

        if self._status_watcher.server_running:
            self._status_watcher._server_unregistering = True

            def _on_complete(server):
                if self.addr.proto == 'tcp' and self.addr.is_connected:
                    self.addr.disconnect()
                self._status_watcher._server_unregistering = False
                if self._pid is not None:
                    self._server_process.detach()
                    self._server_process = None
                    self._pid = None
                fn.value(on_complete, self)
                _libsc3.main._atexitq.remove(self._unregister_atexit)

            def _on_failure(server):
                self._status_watcher._server_unregistering = False
                fn.value(on_failure, self)

            self._status_watcher._add_action(
                'unregister', _on_complete, _on_failure)
            self._status_watcher._unregister()
        else:
            _logger.info(f"server '{self.name}' is not registered")

    def _unregister_atexit(self):
        if self._status_watcher.server_running:
            self._status_watcher._unregister()
            # self.send_msg('/notify', 0, self.client_id)
            # _logger.info(f"server '{self.name}' requested id unregistration")

    def boot(self, register=True, on_complete=None, on_failure=None):
        if _libsc3.main is _libsc3.NrtMain:
            if self._status_watcher.server_running\
            or self._status_watcher.unresponsive:
                _logger.error("can't use an rt server in nrt")
            self._status_watcher._boot_nrt()
            return

        if self._status_watcher.unresponsive:
            _logger.info(f"server '{self.name}' unresponsive, rebooting...")
            self.quit(False)

        if self._status_watcher.server_running:
            _logger.info(f"server '{self.name}' already running")
            return

        if self._status_watcher._server_booting:
            _logger.info(f"server '{self.name}' already booting")
            return

        self._status_watcher._server_booting = True

        def _on_complete(server):
            server._status_watcher._server_booting = False
            server._status_watcher._server_rebooting = False
            server._boot_init()
            fn.value(on_complete, server)
            _libsc3.main._atexitq.add(
                _libsc3.main._atexitprio.SERVERS, self._quit_atexit)

        def _on_failure(server):
            if self.addr.proto == 'tcp':
                self.addr.disconnect()
            server._status_watcher._server_booting = False
            server._status_watcher._server_rebooting = False
            fn.value(on_failure, server)

        self._status_watcher._add_action('boot', _on_complete, _on_failure)

        if not self._is_local:
            _logger.info(f"remote server '{self.name}' needs manual boot")
        else:
            def boot_task():
                # Server can be quitting and booting at the same time, e.g.
                # rebooting, pid release blocks boot until quit is completed.
                if self._pid is not None:  # Is or was running.
                    yield from self._pid_release_condition.hang()

                if self._in_process:
                    _logger.info('booting internal server')
                    self._boot_in_process()  # Not implemented yet.
                    self._pid = _libsc3.main.pid  # Not implemented yet.
                else:
                    self._disconnect_shm()  # Not implemented yet.
                    try:
                        self._server_process = ServerProcess(
                            self._on_server_process_exit)
                        self._server_process.run(self)
                    except OSError as e:
                        self._server_process = None
                        self._status_watcher._server_booting = False
                        self._status_watcher._server_rebooting = False
                        if e.errno == errno.ENOENT\
                        and e.filename == self.options.program:
                            _logger.error(
                                f'{self.name} failed to boot, program '
                                f'{e.filename} not found')
                            return
                        else:
                            raise
                    self._pid = self._server_process.proc.pid
                    _logger.info(
                        f"booting server '{self.name}' on address "
                        f"{self.addr.hostname}:{self.addr.port}")

                if register:
                    # Needs delay to avoid registering to another local server
                    # from another client in case of address already in use.
                    if self.options.protocol == 'tcp' and not self._in_process:
                        def success():
                            self._status_watcher._start_alive_thread(1)
                        def failure():
                            self._status_watcher._server_booting = False
                            self._status_watcher._server_rebooting = False
                            self._status_watcher._clear_actions()
                        self.addr.connect(success, failure)
                    else:
                        self._status_watcher._start_alive_thread(1)

            stm.Routine.run(boot_task, clk.AppClock)

    def _boot_init(self):
        if self.dump_mode != 0:
            self.send_msg('/dumpOSC', self.dump_mode)
        self._connect_shm()  # BUG: not implemented yet

    def _on_server_process_exit(self, exit_code):
        # This method is called after quit or crash.
        self._server_process = None
        self._pid = None
        self._pid_release_condition.signal()
        _logger.info(f"server '{self.name}' exited with exit code {exit_code}")
        if self._status_watcher._server_booting\
        and not self._status_watcher._server_rebooting:
            self._status_watcher._server_booting = False
            self._status_watcher._perform_actions('boot', 'on_failure')
            _logger.warning(f"server '{self.name}' failed to boot")
        # In case of server crash or Exception in World_OpenUDP: bind: Address
        # already in use, status_watcher should stop to avoid registering to
        # another local server. See note in boot(). There sould be a better
        # option for the second case to avoid delay which is arbitrary.
        self._status_watcher._quit(False)

    def _quit_atexit(self):
        if self._status_watcher.server_running:
            event = threading.Event()
            set_func = lambda: event.set()
            self.quit(True, set_func, set_func)
            event.wait(5)  # _MainThread, set_func runs in AppClock thread.

    def reboot(self, func=None, on_failure=None):
        if _libsc3.main is _libsc3.NrtMain:
            _logger.warning("can't reboot a nrt server")
            return

        # // func is evaluated when server is off.
        if not self._is_local:
            _logger.info("can't reboot a remote server")
            return

        if self._status_watcher._server_booting:
            _logger.info(f"server '{self.name}' already booting")
            return

        self._status_watcher._server_rebooting = True

        if self._status_watcher.server_running\
        and not self._status_watcher.unresponsive:
            def _on_complete():
                if func is not None:
                    func()
                self.boot()
            self.quit(True, _on_complete, on_failure)
        else:
            if func is not None:
                func()
            self.boot(on_failure=on_failure)

    @property
    def program_running(self):
        if self._server_process is None:
            return False
        return self._server_process.running()

    def quit(self, watch_shutdown=True, on_complete=None, on_failure=None):
        if _libsc3.main is _libsc3.NrtMain:
            self._status_watcher._quit_nrt()
            return

        # if server is not running or is running but unresponsive.
        if not self._status_watcher.server_running\
        or self._status_watcher.unresponsive:
            _logger.info(f"server '{self.name}' is not running")
            return

        if self._status_watcher._server_quitting:
            _logger.info(f"server '{self.name}' already quitting")
            return

        self._status_watcher._server_quitting = True

        def _on_complete():
            if self.addr.proto == 'tcp':
                self.addr.disconnect()
            self._status_watcher._server_quitting = False
            _libsc3.main._atexitq.remove(self._quit_atexit)
            fn.value(on_complete, self)

        def _on_failure():
            self._status_watcher._server_quitting = False
            fn.value(on_failure, self)

        if watch_shutdown and self._status_watcher.unresponsive:
            _logger.info(
                f"server '{self.name}' was unresponsive, quitting anyway")
            watch_shutdown = False

        self._status_watcher._add_action('quit', _on_complete, _on_failure)
        self._status_watcher._quit(watch_shutdown)
        self.addr.send_msg('/quit')  # Send quit after responders are in place.

        if self._in_process:
            self._quit_in_process()  # Not implemented.
            _logger.info('internal server has quit')
        else:
            _logger.info(f"'/quit' message sent to server '{self.name}'")

        # if(scopeWindow.notNil) { scopeWindow.quit }  # No GUI.
        self._volume.free_synth()
        nod.RootNode(self).free_all()
        self._set_client_id(0)

    @classmethod
    def quit_all(cls, watch_shutdown=True):
        for server in cls.all:
            if server._is_local:
                server.quit(watch_shutdown)

    @classmethod
    def free_all(cls, even_remote=False):  # All refers to cls.all.
        if even_remote:
            for server in cls.all:
                if server._status_watcher.server_running:
                    server.free_nodes()
        else:
            for server in cls.all:
                if server._is_local and server._status_watcher.server_running:
                    server.free_nodes()

    def free_nodes(self):  # Instance free_all in sclang.
        self.send_msg('/g_freeAll', 0)
        self.send_msg('/clearSched')
        self.init_tree()

    def free_default_group(self):
        self.send_msg('g_freeAll', self._default_group.node_id)

    def free_default_groups(self):
        for group in self._default_groups:
            self.send_msg('g_freeAll', group.node_id)

    @classmethod
    def hard_free_all(cls, even_remote=False):
        if even_remote:
            for server in cls.all:
                server.free_nodes()
        else:
            for server in cls.all:
                if server._is_local:
                    server.free_nodes()


    # L1203
    ### internal server commands ###
    # TODO
    # def _boot_in_process(self):
    # def _quit_in_process(self):

    # L1232
    # /* CmdPeriod support for Server-scope and Server-record and Server-volume */
    # TODO

    def dump_tree(self, controls=False):
        '''
        Ask the server to dump its node tree to stdout.

        Parameters
        ----------
        controls: bool
            If `True` also print synth controls with current values.
        '''

        if not self._is_local:
            _logger.info(f'server {self.name} is not local')
        elif self._pid is None:
            _logger.info(f'server {self.name} is not running')
            return
        nod.RootNode(self).dump_tree(controls)  # Also needs stdout access.

    # L1315
    # funciones set/getControlBug*
    # TODO

    def input_bus(self):  # utility
        return bus.AudioBus(
            self.options.input_channels,
            self, self.options.output_channels)

    def output_bus(self):  # utility
        return bus.AudioBus(self.options.output_channels, self, 0)


    ### Node parameter interface ###

    def _as_target(self):
        return self._default_group

    # def scsynth(cls): No.
    # def supernova(cls): No.
    # def from_name(cls): No.
    # def kill_all(cls): No.



"""
Support for built in or extensions data types as UGen or Node
parameters. This module is an internal interface. To extend type
support classes inherit either from UGenParameter or NodeParameter
or both, define _param_type() for the extending type and override
the needed interface methods.

UGenParameter and NodeParameter are values intended for UGens parameters
and Node messages respectively. UGen, Node, SynthDef and SynthDescLib
classes make use of this interface to prepare data in the right format.
This functionality shoudln't be exposed to the user interface, supported
data types must just work with its own interface.
"""

from math import isnan

from ..base import classlibrary as clb
from ..base import utils as utl
from . import _specialindex as _si
from . import _fmtrw as frw


clb.ClassLibrary.late_imports(__name__,
    ('sc3.synth.server', 'srv'),
    ('sc3.synth.node', 'nod'),
    ('sc3.synth.ugens.line', 'lne')
)


### Graphs Parameter Base Class ###

class GraphParameter():
    def __init__(self, value):
        self.__param_value = value

    @property
    def _param_value(self):
        return self.__param_value

    @classmethod
    def _param_type(cls):
        return (cls,)

    def __repr__(self):
        if self is self._param_value:
            return super().__repr__()   # *** BUG: MRO/MULTIPLE INHERITANCE
        else:
            return f'{type(self).__name__}({repr(self._param_value)})'


### UGen graph parameter interface ###

class UGenParameter(GraphParameter):
    def _is_valid_ugen_input(self):
        return False

    def _as_ugen_input(self, *_):
        return self._param_value

    def _as_audio_rate_input(self):
        if self._as_ugen_rate() != 'audio':
            return lne.K2A.ar(self._param_value)
        else:
            return self._param_value

    def _as_ugen_rate(self):  # Was rate for many non UGen objects in sclang.
        try:
            return self.rate
        except AttributeError as e:
            raise AttributeError(
                f'{type(self).__name__} does not implement '
                'rate attribute or _as_ugen_rate method') from e

    def _perform_binary_op_on_ugen(self, selector, ugen):
        selector = _si.sc_opname(selector.__name__)
        raise TypeError(
            f"operation '{selector}' is not supported between "
            f"UGen and {type(self._param_value).__name__}")

    def _r_perform_binary_op_on_ugen(self, selector, ugen):
        selector = _si.sc_opname(selector.__name__)
        raise TypeError(
            f"operation '{selector}' is not supported between "
            f"{type(self._param_value).__name__} and UGen")

    def _write_input_spec(self, file, synthdef):
        raise NotImplementedError(
            f'{type(self).__name__} does '
            'not implement _write_input_spec()')


class UGenNone(UGenParameter):
    @classmethod
    def _param_type(cls):
        return (type(None),)

    def _as_ugen_rate(self):
        return None


class UGenString(UGenParameter):
    @classmethod
    def _param_type(cls):
        return (str,)

    def _as_ugen_rate(self):
        return 'scalar'


class UGenScalar(UGenParameter):
    @classmethod
    def _param_type(cls):
        return (int, float, bool)

    def _is_valid_ugen_input(self):
        return not isnan(self._param_value)

    def _as_audio_rate_input(self):
        if self._param_value == 0:
            return lne.Silent.ar()
        else:
            return lne.DC.ar(self._param_value)

    def _as_ugen_rate(self):
        return 'scalar'

    def _write_input_spec(self, file, synthdef):
        try:
            const_index = synthdef._constants[float(self._param_value)]
            frw.write_i32(file, -1)
            frw.write_i32(file, const_index)
        except KeyError as e:
            raise Exception(
                '_write_input_spec constant not found: '
                f'{float(self._param_value)}') from e


    ### UGen convenience methods (keep in sync) ###
    # These methods are only for ChannelList _multichannel_perform (multichannel
    # expansion UGen compatibility). Some of them could be used in different
    # places within ugens however I prefer to avoid its use. For them to work
    # consistently ugens inputs should be stored as UGenParameter instances and
    # I'm reluctant to do that at least by now, mostly because UGenScalar is
    # the only type compatible with signal operations and graph parameters are
    # being used as input translators so far.

    # def dup(self, n=2):  # Function and Object.
    # def madd(self, mul=1.0, add=0.0):  # Array, SimpleNumber and UGen.
    # def range(self, lo=0.0, hi=1.0):  # SimpleNumber (wslib!) and UGen. # *** BUG: two signatures in builtins
    # def exprange(self, lo=0.01, hi=1.0):  # SequenceableCollection and UGen (and Env).
    # def curverange(self, lo=0.0, hi=1.0, curve=-4):  # SequenceableCollection and UGen (and Env).
    # def unipolar(self, mul=1):  # SequenceableCollection and UGen.
    # def bipolar(self, mul=1):  # SequenceableCollection and UGen.

    def clip(self, lo=0.0, hi=1.0):
        return bi.clip(self._param_value, lo, hi)

    def fold(self, lo=0.0, hi=1.0):  # *** BUG: two signatures in builtins
        return bi.fold(self._param_value, lo, hi)

    def wrap(self, lo=0.0, hi=1.0):  # *** BUG: two signatures in builtins
        return bi.wrap(self._param_value, lo, hi)

    # def min_nyquist(self):  # SequenceableCollection and UGen.

    def degrad(self):
        return bi.degrad(self._param_value)

    def raddeg(self):
        return bi.raddeg(self._param_value)

    def blend(self, other, frac=0.5):
        return bi.blend(self._param_value, other, frac)

    def lag(self, time=0.1):  # SimpleNumber (^this) SequenceableCollection and UGen, idem until prune.
        return self._param_value

    def lag2(self, time=0.1):
        return self._param_value

    def lag3(self, time=0.1):
        return self._param_value

    def lagud(self, utime=0.1, dtime=0.1):
        return self._param_value

    def lag2ud(self, utime=0.1, dtime=0.1):
        return self._param_value

    def lag3ud(self, utime=0.1, dtime=0.1):
        return self._param_value

    def varlag(self, time=0.1, curvature=0, wrap=5, start=None):
        return self._param_value

    def slew(self, up=1, down=1):
        return self._param_value

    def prune(self, min, max, type='minmax'):
        return self._param_value

    # snap is not implemented
    # softround is not implemented

    def linlin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return bi.linlin(self._param_value, inmin, inmax, outmin, outmax, clip)

    def linexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return bi.linexp(self._param_value, inmin, inmax, outmin, outmax, clip)

    def explin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return bi.explin(self._param_value, inmin, inmax, outmin, outmax, clip)

    def expexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return bi.expexp(self._param_value, inmin, inmax, outmin, outmax, clip)

    def lincurve(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        return bi.lincurve(
            self._param_value, inmin, inmax, outmin, outmax, curve, clip)

    def curvelin(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        return bi.curvelin(
            self._param_value, inmin, inmax, outmin, outmax, curve, clip)

    def bilin(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        return bi.bilin(
            self._param_value, incenter, inmin, inmax,
            outcenter, outmin, outmax, clip)

    def biexp(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        return bi.biexp(
            self._param_value, incenter, inmin, inmax,
            outcenter, outmin, outmax, clip)

    def moddif(self, that=0.0, mod=1.0):
        return bi.moddif(self._param_value, that, mod)


class UGenSequence(UGenParameter):
    @classmethod
    def _param_type(cls):
        # tuple prevents multichannel expansion,
        # type must be kept for UGen inputs.
        return (list, tuple)

    def _is_valid_ugen_input(self):
        return True if self._param_value else False  # *** BUG: en sclang, debería comprobar isEmpty porque tira error en SynthDesc (SinOsc.ar() * []). O tal vez cuando construye BinaryOpUGen? Ver el grafo generado, tal vez deba ser None.

    def _as_ugen_input(self, *ugen_cls):
        m = map(
            lambda x: ugen_param(x)._as_ugen_input(*ugen_cls),
            self._param_value)
        return type(self._param_value)(m)

    def _as_audio_rate_input(self, *ugen_cls):
        m = map(
            lambda x: ugen_param(x)._as_audio_rate_input(*ugen_cls),
            self._param_value)
        return type(self._param_value)(m)

    def _as_ugen_rate(self):
        if len(self._param_value) == 1:
            return ugen_param(self._param_value[0])._as_ugen_rate()
        else:
            return utl.list_min(
                [ugen_param(item)._as_ugen_rate() or 'scalar'
                for item in self._param_value])

    def _write_input_spec(self, file, synthdef):
        for item in self._param_value:
            ugen_param(item)._write_input_spec(file, synthdef)


### Node Graph Parameters ###

class NodeParameter(GraphParameter):
    # asTarget.sc interface
    def _as_target(self):
        raise TypeError(
            'invalid value for Node target: '
            f'{type(self._param_value).__name__}')

    def _as_control_input(self):
        return self._param_value

    # extConvertToOSC.sc interface
    # // The following interface in an optimized version of asControlInput that
    # // flattens arrayed values and marks the outermost array of a value with
    # // '[' and ']'. These Chars are turning into typetags ([ and ]) in the OSC
    # // message to mark that array. Inner arrays are flattened (they are not
    # // meaningful in the server context), this makes it possible to write
    # // Synth("test", [0, [[100,200,300], [0.1,0.2,0.3], [10,20,30]] ]) and
    # // have all the arguments be assigned to consecutive controls in the synth.

    def _as_osc_arg_list(self):
        return [self._as_control_input()]

    def _embed_as_osc_arg(self, lst):
        lst.append(self._as_control_input())


class NodeNone(NodeParameter):
    @classmethod
    def _param_type(cls):
        return (type(None),)

    def _as_target(self):
        return srv.Server.default.default_group


class NodeScalar(NodeParameter):
    @classmethod
    def _param_type(cls):
        return (int, float)

    def _as_target(self):
        return nod.Group.basic_new(srv.Server.default, self._param_value)


class NodeString(NodeParameter):
    @classmethod
    def _param_type(cls):
        return (str,)


class NodeSequence(NodeParameter):
    @classmethod
    def _param_type(cls):
        return (list, tuple)

    def _as_control_input(self):
        return type(self._param_value)(
            node_param(x)._as_control_input() for x in self._param_value)

    def _as_osc_arg_list(self):
        lst = []
        for e in self._param_value:
            node_param(e)._embed_as_osc_arg(lst)
        return lst

    def _embed_as_osc_arg(self, lst):
        lst.append('[')
        for e in self._param_value:
            node_param(e)._embed_as_osc_arg(lst)
        lst.append(']')


class NodeDictionary(NodeParameter):
    # Used to convert dict to a flat list of key-value pairs.

    @classmethod
    def _param_type(cls):
        return (dict,)

    def _as_control_input(self):
        return list(
            node_param(x)._as_control_input()
            for items in self._param_value.items() for x in items)

    def _as_osc_arg_list(self):
        return self._as_control_input()

    def _embed_as_osc_arg(self, lst):
        lst.append('[')
        for item in self._param_value.items():
            for e in item:
                node_param(e)._embed_as_osc_arg(lst)
        lst.append(']')


### Module functions ###

def _graph_param(obj, param_cls):
    new_cls = None
    for sub_class in param_cls.__subclasses__():
        if isinstance(obj, sub_class._param_type()):
            new_cls = sub_class
            break
    if new_cls is None:
        raise TypeError(
            f"{param_cls.__name__}: type '{type(obj).__name__}' not supported")
    return new_cls(obj)


def ugen_param(obj):
    if isinstance(obj, UGenParameter):
        return obj
    return _graph_param(obj, UGenParameter)


def node_param(obj):
    if isinstance(obj, NodeParameter):
        return obj
    return _graph_param(obj, NodeParameter)



'''
Event.sc attempt 3.2. No multichannel expasion.
'''

import types
import sys

from ..base import builtins as bi
from ..base import operand as opd
from ..synth import server as srv
from ..synth import synthdesc as sdc
from ..synth import node as nod
from ..synth import _graphparam as gpp
from . import scale as scl


__all__ = ['event', 'new_event', 'Rest', 'silent', 'is_rest']


### Rest ###


class Rest(opd.Operand):
    def __init__(self, value=1.0):
        super().__init__(value)

    def __bool__(self):
        # unwrapBoolean
        return self.value


def silent(dur=1.0, inevent=None):
    if inevent is None:
        inevent = event()
    else:
        inevent = inevent.copy()
    inevent['delta'] = dur * inevent.get('stretch', 1.0)
    inevent['dur'] = dur if isinstance(dur, Rest) else Rest(dur)
    return inevent


def is_rest(inevent):
    return (inevent.get('type') == 'rest' or
            any(isinstance(value, Rest) for value in inevent.values()))


### Event Keys ###


class keyfunction():
    '''Decorator class as mark for instance methods that become keys.'''

    def __init__(self, func):
        self.func = func


class _PrepareDict(dict):
    '''
    Especial dictionary to collect attributes and methods. It allows
    to repeat names but classes can't have public value-attributes.
    '''

    def __init__(self):
        super().__setitem__('default_values', dict())
        super().__setitem__('default_functions', dict())

    def __setitem__(self, key, value):
        if key.startswith('_') or isinstance(value, types.FunctionType):
            super().__setitem__(key, value)
        elif isinstance(value, keyfunction):
            self['default_functions'][key] = value.func
        else:
            self['default_values'][key] = value


class MetaEventDict(type):
    _event_types = dict()

    @classmethod
    def __prepare__(meta_cls, cls_name, args, **kwargs):
        pd = _PrepareDict()
        if 'partial_events' in kwargs:
            # Inherit non-key object attributes.
            dk = ('default_values', 'default_functions')
            for pe in kwargs['partial_events']:
                ed = vars(pe)
                keys = ed.keys() - dk
                for k in keys:
                    if k.startswith('__'):
                        pass
                    else:
                        pd[k] = ed[k]
        return pd

    def __init__(cls, name, bases, cls_dict, partial_events=None):
        default_values = dict()
        default_functions = dict()
        if name == 'EventDict':
            return
        if partial_events is not None:
            for pe in partial_events:  # reversed?
                default_values.update(pe.default_values)
                default_functions.update(pe.default_functions)
        for base in reversed(bases):
            if base is EventDict:
                break
            if isinstance(base, MetaEventDict):
                default_values.update(base.default_values)
                default_functions.update(base.default_functions)
        for key, value in cls.default_values.items():
            default_values[sys.intern(key)] = value
        for key, value in cls.default_functions.items():
            default_functions[sys.intern(key)] = value
        cls.default_values = default_values
        cls.default_functions = default_functions
        if 'type' in cls.default_values\
        and cls.default_values['type'] is not None:
            cls._event_types[cls.default_values['type']] = cls

    @property
    def types(cls):
        return tuple(cls._event_types.keys())


class EventDict(dict, metaclass=MetaEventDict):
    def __init_subclass__(cls, partial_events=None, **kwargs):
        super().__init_subclass__(**kwargs)

    def __call__(self, key, *args):
        if key in self.default_functions:
            return self.default_functions[key](self, *args)
        if not key in self:
            return self.default_values[key]
        if callable(self[key]):
            return self[key](self, *args)
        else:
            return self[key]

    def __copy__(self):
        return self.copy()

    def copy(self):
        return type(self)(self)

    def __repr__(self):
        return f'{type(self).__name__}({super().__repr__()})'


def new_event(name, values=None, functions=None,
              bases=None, partial_events=None):
    def init_defaults(ns):
        if values is not None:
            for k, v in values.items():
                ns[k] = v
        if functions is not None:
            for k, v in functions.items():
                if not isinstance(v, types.FunctionType):
                    raise ValueError(f'{v} is not FunctionType')
                ns[k] = keyfunction(v)

    if bases is None:
        bases = (EventDict,)
    if partial_events is None:
        partial_events = {'partial_events': ()}
    else:
        partial_events = {'partial_events': partial_events}
    return types.new_class(name, bases, partial_events, init_defaults)


class event(EventDict):
    _default_type = 'note'

    def __new__(cls, *args, **kwargs):
        d = {**dict(*args), **kwargs}  # Override duplicated 'type' keys.
        type = d.pop('type', None)  # Remove 'type' from actual keys.
        if type:
            try:
                return cls._event_types[type](d)
            except KeyError:
                raise ValueError(f"no event type '{type}'") from None
        else:
            return cls._event_types[cls._default_type](d)


### Partial Events ###


class PartialEvent(EventDict):
    pass


class PitchKeys(PartialEvent):
    freq = bi.midicps(60)
    detune = 0.0
    harmonic = 1.0

    midinote = 60
    ctranspose = 0.0

    # note = 0  # Desn't need a default, see comment in keyfunction.
    degree = 0
    mtranspose = 0
    gtranspose = 0.0
    octave = 5.0
    root = 0.0
    scale = scl.Scale([0, 2, 4, 5, 7, 9, 11])

    def _detuned_freq(self):
        return self('freq') * self('harmonic') + self('detune')

    @keyfunction
    def freq(self):
        if 'freq' in self:
            return self['freq']
        elif 'midinote' in self or 'note' in self:
            return self._freq_from_midinote()
        elif 'degree' in self:
            return self._freq_from_degree()
        else:
            return self.default_values['freq']

    def _freq_from_midinote(self):
        return bi.midicps(self._transposed_midinote())

    def _freq_from_degree(self):
        # No ctranspose. For every main key with it own modifiers (harmonic
        # and detune for freq) the other two can be calculate (without their
        # modifiers). Only one main key should be used at a time and 'freq'
        # overrides 'midinote' that overrides 'degree' for events.
        return bi.midicps(self._midinote_from_degree())

    def _transposed_midinote(self):
        return self('midinote') + self('ctranspose')

    @keyfunction
    def midinote(self):
        if 'midinote' in self:
            return self['midinote']
        elif 'note' in self:
            return self._midi_from_note()
        elif 'degree' in self:
            return self._midinote_from_degree()
        elif 'freq' in self:
            return self._midinote_from_freq()
        else:
            return self.default_values['midinote']

    def _midi_from_note(self):
        # See comment in keyfunction.
        ret = self['note'] + self('gtranspose') + self('root')
        ret = ret / self('scale').tuning.spo + self('octave') - 5.0
        ret = ret * (12.0 * bi.log2(self('scale').tuning.octave_ratio)) + 60
        return ret

    def _midinote_from_degree(self):
        scale = self('scale')
        ret = scale.degree_to_key(self('degree') + self('mtranspose'))
        ret = ret + self('gtranspose') + self('root')
        ret = ret / scale.tuning.spo + self('octave') - 5.0
        ret = ret * (12.0 * bi.log2(scale.tuning.octave_ratio)) + 60
        return ret

    def _midinote_from_freq(self):
        return bi.cpsmidi(self._detuned_freq())

    @keyfunction
    def note(self):
        # When set, this key doesn't call degree_to_key when converting to
        # midinote so it can be done externally, yet this is not the best
        # path for combinations and naming/meaning gets confusing.
        return self('scale').degree_to_key(self('degree') + self('mtranspose'))

    @keyfunction
    def degree(self):
        if 'degree' in self:
            return self['degree']
        elif 'freq' in self:
            return self._degree_from_freq()
        elif 'midinote' in self:
            return self._degree_from_midinote()
        else:
            return self.default_values['degree']

    def _degree_from_freq(self):
        return self._midinote_to_degree(bi.cpsmidi(self._detuned_freq()))

    def _degree_from_midinote(self):
        return self._midinote_to_degree(self._transposed_midinote())

    def _midinote_to_degree(self, midinote):
        # From SequenceableCollection.performDegreeToKey
        scale = self('scale')
        degree_root = (midinote // 12 - 5) * len(scale.tuning)
        key = midinote % 12
        # From SequenceableCollection.indexInBetween modified.
        index = None
        for i, value in enumerate(scale.tuning):
            if value > key:
                index = i
                break
        if index == 0:
            return 0.0
        if index is None:
            index = i + 1
            a = scale.tuning[-1]
            b = scale.tuning[0] + scale.tuning.spo
        else:
            a = scale.tuning[index - 1]
            b = scale.tuning[index]
        div = b - a
        if div == 0:
            return float(index)
        else:
            return (key - a) / div + index - 1 + degree_root


class DurationKeys(PartialEvent):
    delta = None  # Used in EventStreamPlayer._play_and_delta.
    sustain = None
    dur = 1.0
    legato = 0.8
    stretch = 1.0

    # tempo = None
    # lag = 0.0

    # strum = 0.0
    # strum_ends_together = False

    @keyfunction
    def delta(self):
        # NOTE: Cast from Rest is done externally (explicit).
        if 'delta' in self:
            return self['delta']
        else:
            return self('dur') * self('stretch')

    @keyfunction
    def sustain(self):
        if 'sustain' in self:
            return self['sustain']
        else:
            return self('dur') * self('legato') * self('stretch')

    # *** Behaviour is still missing for some keys.


class AmplitudeKeys(PartialEvent):
    amp = 0.1
    db = -20
    velocity = 12  # 0.1amp == -20dB == 12vel, linear mapping (-42.076dB range for velocity)

    pan = 0.0
    trig = 0.5   # Why was?

    @keyfunction
    def amp(self):
        if 'amp' in self:
            return self['amp']
        elif 'db' in self:
            return bi.dbamp(self['db'])
        elif 'velocity' in self:
            return self._amp_from_velocity()
        else:
            return self.default_values['amp']

    def _amp_from_velocity(self):
        return self['velocity'] / 127

    @keyfunction
    def db(self):
        if 'db' in self:
            return self['db']
        elif 'amp' in self:
            return bi.ampdb(self['amp'])
        elif 'velocity' in self:
            return self._db_from_velocity()
        else:
            return self.default_values['db']

    def _db_from_velocity(self):
        return bi.ampdb(self._amp_from_velocity())

    @keyfunction
    def velocity(self):
        if 'velocity' in self:
            return self['velocity']
        elif 'amp' in self:
            return self._velocity_from_amp()
        elif 'db' in self:
            return self._velocity_from_db()
        else:
            return self.default_values['velocity']

    def _velocity_from_amp(self):
        return int(127 * self['amp'])

    def _velocity_from_db(self):
        return int(127 * bi.dbamp(self['db']))


class ServerKeys(PartialEvent):
    server = None
    latency = None
    node_id = None
    group = None
    synth_lib = None
    synth_desc = None
    out = 0
    add_action = 'addToHead'
    msg_params = []
    instrument = 'default'
    variant = None
    has_gate = True  # // assume SynthDef has gate
    send_gate = None  # // sendGate == false turns off releases
    args = ('freq', 'amp', 'pan', 'trig')  # // for 'type' 'set'
    # lag = 0
    timing_offset = 0  # Used in EventStreamPlayer._synch_with_quant

    @keyfunction
    def server(self):
        return srv.Server.default

    @keyfunction
    def group(self):
        if 'group' in self:
            return self['group']
        else:
            return self('server').default_group.node_id

    @keyfunction
    def synth_lib(self):
        return sdc.SynthDescLib.default  # BUG: Was global_.

    @keyfunction
    def send_gate(self):
        if 'send_gate' in self:
            return self['send_gate']
        else:
            return self('has_gate')

    def _get_msg_params(self):  # Was get_msg_func
        msg_params = self('msg_params')
        if not msg_params or self('is_playing'):
            synth_lib = self('synth_lib')
            desc = synth_lib.at(self('instrument'))
            if desc is None:
                self['msg_params'] = self._default_msg_params()
                return self['msg_params']
            else:
                self['synth_desc'] = desc
                self['has_gate'] = desc.has_gate
                if desc.has_gate and not desc.keep_gate:
                    control_names = desc.control_names[:]
                    control_names.remove('gate')
                else:
                    control_names = desc.control_names
                msg_params = []
                for arg in control_names:
                    if arg in self:
                        msg_params.extend([arg, self(arg)])
                self['msg_params'] = msg_params
                return msg_params
        else:
            return msg_params

    def _default_msg_params(self):  # Was default_msg_func.
        return ['freq', self('freq'), 'amp', self('amp'),
                'pan', self('pan'), 'out', self('out')]

    def _synthdef_name(self):
        if self('variant') is not None\
        and self('synth_desc') is not None\
        and self('synth_desc').has_variants():
            return f"{self('instrument')}.{self('variant')}"
        else:
            return self('instrument')

    # def _sched_bundle(self, lag, offset, server, msg, latency=None):
    #     # // "lag" is a tempo independent absolute lag time (in seconds)
    #     # // "offset" is the delta time for the clock (usually in beats)


# TODO...


### Event Types ###


class EventType(EventDict):
    # type = None
    # is_playing = False

    def play(self):
        pass

    # def __eq__(self, other):
    #     return super().__eq__(other)

    def __hash__(self):
        return id(self)


class NoteEvent(EventType, partial_events=(
        PitchKeys, AmplitudeKeys, DurationKeys, ServerKeys)):  # TODO...
    type = 'note'
    is_playing = False

    def play(self):  # NOTE: No server parameter.
        self['freq'] = self._detuned_freq()  # Before _get_msg_params.
        param_list = self._get_msg_params()  # Populates synth_desc.
        self['instrument'] = instrument = self._synthdef_name()
        self['server'] = server = self('server')

        self['node_id'] = node_id = server.next_node_id()
        add_action = nod.Node.action_number_for(self('add_action'))
        self['group'] = group = gpp.node_param(
            self('group'))._as_control_input()

        bndl = ['/s_new', instrument, node_id, add_action, group]
        bndl.extend(param_list)
        bndl = gpp.node_param(bndl)._as_osc_arg_list()

        # *** BUG: socket.sendto and/or threading mixin use too much cpu.
        server.send_bundle(server.latency, bndl)  # Missing ~latency, ~lag and ~timmingOffset.
        if self('send_gate'):
            server.send_bundle(
                server.latency + self('sustain'),
                ['/n_set', node_id, 'gate', 0])

        self['is_playing'] = True


class _MonoOnEvent(EventType, partial_events=(
        PitchKeys, AmplitudeKeys, DurationKeys, ServerKeys)):
    type = '_mono_on'
    is_playing = False

    def play(self):
        self['add_action'] = nod.Node.action_number_for(self('add_action'))
        self['group'] = gpp.node_param(self('group'))._as_control_input()
        bndl = [
            '/s_new', self['instrument'], self['node_id'],
            self['add_action'], self['group'], *self['msg_params']]
        bndl = gpp.node_param(bndl)._as_osc_arg_list()
        self['server'].send_bundle(self['server'].latency, bndl)  # Missing ~latency, ~lag and ~timmingOffset.
        self['is_playing'] = True

    def _prepare_event(self, instrument):
        self['instrument'] = instrument
        self['freq'] = self._detuned_freq()  # Before _get_msg_params.
        self['msg_params'] = self._get_msg_params()  # Populates synth_desc.
        self['has_gate'] = self('has_gate')
        self['server'] = self('server')
        self['node_id'] = self['server'].next_node_id()


class _MonoSetEvent(EventType, partial_events=(
        PitchKeys, AmplitudeKeys, DurationKeys, ServerKeys)):
    type = '_mono_set'
    is_playing = True
    mono_params = None

    def play(self):
        self['freq'] = self._detuned_freq()
        self['server'] = self('server')
        bndl = ['/n_set', self['node_id'], *self._update_msg_params()]
        bndl = gpp.node_param(bndl)._as_osc_arg_list()
        self['server'].send_bundle(self['server'].latency, bndl)

    def _update_msg_params(self):
        msg_params = []
        for arg in self['mono_params']:
            msg_params.extend([arg, self(arg)])
        self['msg_params'] = msg_params
        return msg_params


class _MonoOffEvent(EventType, partial_events=(ServerKeys,)):
    type = '_mono_off'
    is_playing = True
    has_gate = False
    gate = 0
    delay = 0

    def play(self):
        if self('has_gate'):
            bndl = ['/n_set', self['node_id'], 'gate', self('gate')]
        else:
            bndl = ['/n_free', self['node_id']]
        bndl = gpp.node_param(bndl)._as_osc_arg_list()
        server = self('server')
        server.send_bundle(server.latency + self('delay'), bndl)
        self['is_playing'] = False



"""FilterPatterns.sc"""

import copy
import logging

from ...base import builtins as bi
from ...base import stream as stm
from ...base import utils as utl
from ...base import functions as fn
from .. import pattern as ptt
from .. import event as evt


class FilterPattern(ptt.Pattern):
    def __init__(self, pattern):
        self.pattern = pattern
        self._is_event_pattern = pattern.is_event_pattern

    @property
    def is_event_pattern(self):
        return self._is_event_pattern


class Pn(FilterPattern):
    # NOTE: For Pn to have 'key' and to be parent of Pgate seems to be a bit
    # arbitrary. A gate is not usually something that repeats n times, moreover
    # Pgate acts like a hold more that a gate.
    def __init__(self, pattern, repeats=float('inf'), key=None):
        super().__init__(pattern)
        self.repeats = repeats
        self.key = key

    def __embed__(self, inevent):
        pattern = self.pattern
        key = self.key
        if key is None:
            for _ in bi.counter(self.repeats):
                inevent = yield from pattern.__embed__(inevent)
        else:
            for _ in bi.counter(self.repeats):
                inevent[key] = True
                inevent = yield from pattern.__embed__(inevent)
            inevent[key] = False
        return inevent

    # storeArgs


class Pgate(Pn):
    def __embed__(self, inevent):
        pattern = self.pattern
        key = self.key
        stream = output = None
        for _ in bi.counter(self.repeats):
            stream = stm.stream(pattern)
            try:
                while True:
                    if inevent.get(key, False) is True or output is None:
                        output = stream.next(inevent)
                    inevent = yield from stm.embed(copy.copy(output), inevent)
            except stm.StopStream:
                pass
            output = None  # // Force new value for every repeat.
        return inevent

    # storeArgs


class FuncFilterPattern(FilterPattern):
    def __init__(self, func, pattern):
        super().__init__(pattern)
        self.func = func

    # storeArgs


class Pcollect(FuncFilterPattern):
    def __embed__(self, inval):
        func = self.func
        stream = stm.stream(self.pattern)
        outval = None
        try:
            while True:
                outval = stream.next(inval)
                inval = yield fn.value(func, outval, inval)
        except stm.StopStream:
            pass
        return inval

    # asStream  # For some reason it converts to a FunctonStream.


class Pselect(FuncFilterPattern):
    def __embed__(self, inval):
        func = self.func
        stream = stm.stream(self.pattern)
        outval = None
        try:
            while True:
                outval = stream.next(inval)
                if fn.value(func, outval, inval) is True:
                    inval = yield outval
        except stm.StopStream:
            pass
        return inval

    # asStream  # Idem.


class Preject(FuncFilterPattern):
    def __embed__(self, inval):
        func = self.func
        stream = stm.stream(self.pattern)
        outval = None
        try:
            while True:
                outval = stream.next(inval)
                if fn.value(func, outval, inval) is False:
                    inval = yield outval
        except stm.StopStream:
            pass
        return inval

    # asStream  # Idem.


# These are all specific variants of Pchain + Pcollect.
# class Pfset(FuncFilterPattern): ...
# class Psetpre(FilterPattern): ...
# class Paddpre(Psetpre): ...
# class Pmulpre(Psetpre): ...
# class Pset(FilterPattern): ...
# class Padd(Pset): ...
# class Pmul(Pset): ...
# class Psetp(Pset): ...
# class Paddp(Psetp): ...
# class Pmulp(Psetp): ...
# class Pstretch(FilterPattern): ...
# class Pstretchp(Pstretch): ...
# class Pbindf(FilterPattern): ...


# class Pplayer(FilterPattern):  # Undocumented, with note.


class Pdrop(FilterPattern):
    def __init__(self, pattern, n):
        super().__init__(pattern)
        self.n = int(n)

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        first_inval = inval
        try:
            for _ in range(self.n):
                inval = stream.next(first_inval)
            while True:
                inval = yield stream.next(inval)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Plen(FilterPattern):  # Was Pfin.
    def __init__(self, pattern, n):
        super().__init__(pattern)
        self.n = int(n)

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        try:
            for _ in range(self.n):
                inval = yield stream.next(inval)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pdur(ptt.EventPattern, FilterPattern):  # Was Pfindur.
    def __init__(self, pattern, dur, tolerance=0.001, quant=None):
        super(ptt.EventPattern, self).__init__(pattern)
        self.dur = dur
        self.tolerance = tolerance
        self.quant = quant

    def __embed__(self, inevent):
        elapsed = 0.0
        local_dur = self.dur
        tolerance = self.tolerance
        quant = self.quant
        stream = stm.stream(self.pattern)
        delta = next_elapsed = remaining = None
        try:
            while True:
                inevent = stream.next(inevent)
                delta = inevent('delta')
                next_elapsed = elapsed + float(delta)
                if bi.roundup(next_elapsed, tolerance) >= local_dur:
                    remaining = local_dur - elapsed
                    inevent = inevent.copy()
                    inevent['delta'] = type(delta)(remaining)
                    return (yield inevent)
                elapsed = next_elapsed
                inevent = yield inevent
        except stm.StopStream:
            if quant is not None:
                delta = bi.roundup(elapsed, quant) - elapsed
                if delta > 0:
                    inevent = yield evt.silent(delta, inevent)
        return inevent

    # storeArgs


# class Psync(FilterPattern):  # Added as quant to Pdur.


class Pconst(FilterPattern):
    def __init__(self, pattern, sum, tolerance=0.001):
        super().__init__(pattern)
        self.sum = sum
        self.tolerance = tolerance

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        local_sum = self.sum
        tolerance = self.tolerance
        sum = next_sum = 0
        value = None
        try:
            while True:
                value = stream.next(inval)
                next_sum = sum + value
                if bi.roundup(next_sum, tolerance) >= local_sum:
                    inval = yield local_sum - sum
                    return inval
                sum = next_sum
                inval = yield value
        except stm.StopStream:
            inval = yield local_sum - sum
        return inval


class Plag(ptt.EventPattern, FilterPattern):
    def __init__(self, pattern, lag):
        super(ptt.EventPattern, self).__init__(pattern)
        self.lag = lag

    def __embed__(self, inevent):
        yield evt.silent(self.lag, inevent)
        return (yield from stm.embed(self.pattern, inevent))

    # storeArgs


class Pstutter(FilterPattern):
    def __init__(self, pattern, n):
        super().__init__(pattern)
        self.n = n

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        n_stream = stm.stream(self.n)
        value = n = None
        try:
            while True:
                value = stream.next(inval)
                n = n_stream.next(inval)
                for _ in range(abs(n)):
                    inval = yield copy.copy(value)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


# class PdurStutter(Pstutter):  # Tremolo-like partition.


class Platch(FilterPattern):  # Was Pclutch.
    class _UNDEFINED(): pass

    def __init__(self, pattern, trig=True):
        super().__init__(pattern)
        self.trig = trig

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        trig_stream = stm.stream(self.trig)
        UNDEFINED = self._UNDEFINED
        last_inval = UNDEFINED
        trig = None
        try:
            while True:
                trig = trig_stream.next(inval)
                if trig:
                    last_inval = stream.next(inval)
                    inval = yield last_inval
                else:
                    if last_inval is UNDEFINED:
                        last_inval = stream.next(inval)
                    inval = yield copy.copy(last_inval)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pwhile(FuncFilterPattern):
    def __embed__(self, inval):
        func = self.func
        pattern = self.pattern
        while fn.value(func, inval):
            inval = yield from stm.embed(pattern, inval)
        return inval


class Pwrap(FilterPattern):
    def __init__(self, pattern, lo, hi):
        super().__init__(pattern)
        self.lo = lo
        self.hi = hi

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        lo_stream = stm.stream(self.lo)
        hi_stream = stm.stream(self.hi)
        value = lo = hi = None
        try:
            while True:
                lo = lo_stream.next(inval)
                hi = hi_stream.next(inval)
                value = stream.next(inval)
                inval = yield bi.wrap(value, lo, hi)
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Ptrace(FilterPattern):
    _ptrace_logger = logging.getLogger('Ptrace')

    def __init__(self, pattern, prefix=None, keys=None):
        super().__init__(pattern)
        self.prefix = prefix or ''
        self.keys = keys

    def __embed__(self, inval):
        logger = self._ptrace_logger
        stream = stm.stream(self.pattern)
        prefix = self.prefix
        keys = self.keys
        calc_keys = True
        outval = None
        try:
            if self.keys:
                while True:
                    outval = stream.next(inval)
                    if calc_keys:
                        if isinstance(outval, evt.EventType):
                            keys = tuple(
                                k for k in keys if k in outval\
                                or k in outval.default_values)
                        else:
                            keys = tuple()
                        calc_keys = False
                    logger.info('%s%s', prefix, {k: outval(k) for k in keys})
                    inval = yield outval
            else:
                while True:
                    outval = stream.next(inval)
                    logger.info('%s%s', prefix, outval)
                    inval = yield outval
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pclump(FilterPattern):
    def __init__(self, pattern, n):
        super().__init__(pattern)
        self.n = n

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        n_stream = stm.stream(self.n)
        lst = n = value = None
        try:
            while True:
                lst = []
                n = n_stream.next(inval)
                for _ in range(int(n)):
                    value = stream.next(inval)
                    lst.append(value)
                inval = yield lst
        except stm.StopStream:
            if lst:
                inval = yield lst
        return inval

    # storeArgs


class Pflatten(Pclump):
    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        n_stream = stm.stream(self.n)
        flatten = utl.flatten
        n = value = None
        try:
            while True:
                n = n_stream.next(inval)
                value = stream.next(inval)
                if isinstance(value, list):
                    value = flatten([value], n)
                    for item in value:
                        inval = yield item
                else:
                    inval = yield value
        except stm.StopStream:
            pass
        return inval


class Pdiff(FilterPattern):
    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        prev = next = None
        try:
            prev = stream.next(inval)
            while True:
                next = stream.next(inval)
                inval = yield next - prev
                prev = next
        except stm.StopStream:
            pass
        return inval


class Pprorate(FilterPattern):  # Was Protate.
    def __init__(self, pattern, proportion=1):
        super().__init__(pattern)
        self.proportion = proportion

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        prop_stream = stm.stream(self.proportion)
        value = c = None
        try:
            while True:
                value = stream.next(inval)
                c = prop_stream.next(inval)
                if isinstance(c, (list, tuple)):
                    for el in c:
                        inval = yield el * value
                else:
                    inval = yield c * value
                    inval = yield (1 - c) * value
        except stm.StopStream:
            pass
        return inval

    # storeArgs


class Pavaroh(FilterPattern):
    def __init__(self, pattern, aroh, avaroh, spo=12):
        super().__init__(pattern)
        self.aroh = aroh  # Aroh and avaroh are Scale objects.
        self.avaroh = avaroh
        self.spo = spo

    def __embed__(self, inval):
        stream = stm.stream(self.pattern)
        spo_stream = stm.stream(self.spo)
        aroh = self.aroh
        avaroh = self.avaroh
        spo = me = scale = None
        melast = 0
        try:
            while True:
                spo = spo_stream.next(inval)
                me = stream.next(inval)
                scale = aroh if me >= melast else avaroh
                melast = me
                inval = yield scale.degree_to_key(me)
        except stm.StopStream:
            pass
        return inval

    # storeArgs



"""From Patterns.sc & Ppar.sc"""

import logging
import copy

from ...base import stream as stm
from ...base import builtins as bi
from ...base import functions as fn
from ...base import _taskq as tsq
from .. import pattern as ptt
from .. import eventstream as est
from .. import event as evt
from . import listpatterns as lsp


_logger = logging.getLogger(__name__)


### Event patterns ###


class Pkey(ptt.Pattern):
    # Access a key from the input event within a Pbind.
    def __init__(self, key, length=float('inf')):  # Changed: repeats is for lists.
        self.key = key
        self.length = length

    def __embed__(self, inevent):
        key_stream = stm.stream(self.key)
        try:
            for _ in bi.counter(self.length):
                inevent = (yield inevent[key_stream.next(inevent)]) or dict()
        except (stm.StopStream, KeyError):
            pass
        return inevent

    # storeArgs


class Pchain(ptt.EventPattern):
    def __init__(self, *patterns):
        self.patterns = list(patterns)

    def chain(self, pattern):  # <>  # *** NOTE: Maybe a function like stm.stream.
        return type(self)(*self.patterns, pattern)

    def __embed__(self, inevent):
        streams = [stm.stream(p) for p in reversed(self.patterns)]
        try:
            while True:
                inevent = inevent.copy()
                for stream in streams:
                    inevent = stream.next(inevent)
                inevent = yield inevent
        except stm.StopStream:
            pass
        return inevent

    # storeOn


class Pevent(ptt.EventPattern):
    # This class can be used to change the default value of PatternEventStream.
    def __init__(self, pattern, event):
        self.pattern = pattern
        self.event = event

    def __embed__(self, inevent):
        event = self.event
        stream = stm.stream(self.pattern)
        try:
            while True:
                inevent = yield stream.next(event)
        except stm.StopStream:
            pass
        return inevent

    # storeArgs


class Pbind(ptt.EventPattern):
    def __init__(self, mapping):
        self.dict = dict(mapping)

    def __embed__(self, inevent):
        event = None
        stream_dict = {k: stm.stream(v) for k, v in self.dict.items()}

        try:
            while True:
                if inevent is None:
                    return  # Equivalent to ^nil.yield.
                event = inevent.copy()
                event.update(self._stream_dict_next(stream_dict))
                inevent = yield event
        except stm.StopStream:
            pass
        return inevent

    @staticmethod
    def _stream_dict_next(stream_dict):
        event = dict()
        for name, stream in stream_dict.items():
            stream_out = stream.next(event)  # raises StopStream
            if isinstance(name, tuple):
                if not isinstance(stream_out, (list, tuple))\
                or isinstance(stream_out, (list, tuple))\
                and len(name) > len(stream_out):
                    _logger.warning(
                        'the pattern is not providing enough '
                        f'values to assign to the key set: {name}')
                    return inevent
                for i, key in enumerate(name):
                    event[key] = stream_out[i]
            else:
                event[name] = stream_out
        return event

    # storeArgs # TODO


class Pmono(Pbind):
    _kept_keys = {'server', 'node_id', 'has_gate'}

    def __init__(self, synth_name, mapping, articulate=False):
        super().__init__(mapping)
        self.synth_name = synth_name
        self.articulate = articulate

    def __embed__(self, inevent):
        if self.articulate:
            return (yield from self._embed_mono_artic(inevent))
        else:
            return (yield from self._embed_mono(inevent))

    def _embed_mono(self, inevent):
        synth_name = self.synth_name
        kept_keys = self._kept_keys
        stream_dict = {k: stm.stream(v) for k, v in self.dict.items()}
        cleanup = est.CleanupEntry()
        server = node_id = mono_params = event = None

        try:
            while True:
                if node_id is None:
                    event = evt.event(inevent, type='_mono_on')
                    event.update(self._stream_dict_next(stream_dict))
                    event._prepare_event(synth_name)
                    server = event['server']
                    node_id = event['node_id']
                    mono_params = event['msg_params'][::2]  # For _update_msg_params
                    cleanup.add_event(evt.event(
                        {k: event[k] for k in kept_keys}, type='_mono_off'))
                    inevent = yield event
                else:
                    event = evt.event(inevent, type='_mono_set')
                    event.update(self._stream_dict_next(stream_dict))
                    event['server'] = server
                    event['node_id'] = node_id
                    event['mono_params'] = mono_params
                    inevent = yield event
        except stm.StopStream:
            cleanup.run()
        return inevent

    def _embed_mono_artic(self, inevent):
        synth_name = self.synth_name
        kept_keys = self._kept_keys
        stream_dict = {k: stm.stream(v) for k, v in self.dict.items()}
        cleanup = est.CleanupEntry()
        server = node_id = mono_params = event = cleanup_event = None

        try:
            while True:
                if node_id is None:
                    event = evt.event(inevent, type='_mono_on')
                    event.update(self._stream_dict_next(stream_dict))
                    event._prepare_event(synth_name)
                    if event('sustain') >= event('delta')\
                    and not evt.is_rest(event):
                        server = event['server']
                        node_id = event['node_id']
                        mono_params = event['msg_params'][::2]  # For _update_msg_params
                        cleanup_event = evt.event(
                            {k: event[k] for k in kept_keys}, type='_mono_off')
                        cleanup.add_event(cleanup_event)
                        inevent = yield event
                    else:
                        event = evt.event(event, type='note')
                        inevent = yield event
                else:
                    event = evt.event(inevent, type='_mono_set')
                    event.update(self._stream_dict_next(stream_dict))
                    event['server'] = server
                    event['node_id'] = node_id
                    event['mono_params'] = mono_params
                    if event('sustain') < event('delta'):
                        cleanup_event['delay'] = event('sustain')
                        cleanup_event.play()
                        cleanup.remove_event(cleanup_event)
                        node_id = None
                    elif evt.is_rest(event):
                        cleanup_event.play()
                        cleanup.remove_event(cleanup_event)
                        node_id = None
                    inevent = yield event
        except stm.StopStream:
            cleanup.run()
        return inevent


### Ppar.sc ###


class Ppar(ptt.EventPattern):
    def __embed__(self, inevent):
        # assn?
        queue = tsq.TaskQueue()
        for _ in range(self.repeats):
            now = 0.0
            self._init_streams(queue)
            # // If first event not at time zero.
            if not queue.empty():
                nexttime = queue.peek()[0]
                if nexttime > 0.0:
                    outevent = evt.silent(nexttime, inevent)
                    inevent = yield outevent
                    now = nexttime
            while not queue.empty():
                try:
                    stream = queue.pop()[1]
                    outevent = stream.next(inevent)
                    outevent = evt.event(outevent)  # as_event
                    # // Requeue stream.
                    queue.add(now + float(outevent('delta')), stream)
                    nexttime = queue.peek()[0]
                    outevent['delta'] = nexttime - now
                    inevent = yield outevent
                    now = nexttime
                except stm.StopStream:  # next
                    if not queue.empty():
                        # // That child stream ended, so rest until next one.
                        nexttime = queue.peek()[0]
                        outevent = evt.silent(nexttime - now, inevent)
                        inevent = yield outevent
                        now = nexttime
                    else:
                        queue.clear()
            return inevent

    def _init_streams(self, queue):
        for pattern in self.lst:
            queue.add(0.0, stm.stream(pattern))


class Ptpar(Ppar):
    def _init_streams(self, queue):
        for i in range(0, len(self.lst) - 1, 2):  # gen_cclumps(lst, 2)
            queue.add(self.lst[i], stm.stream(self.lst[i + 1]))


class Pgpar(Ppar):
    ...


class Pgtpar(Pgpar):
    ...



"""From Patterns.sc"""


from .. import pattern as ptt


### Imperative patterns ###


class Ptime(ptt.Pattern):
    # // Returns relative time (in beats) from moment of embedding.
    ...


# // if an error is thrown in the stream, func is evaluated
# class Pprotect(FilterPattern): # BUG: FilterPatterns.sc
#     ...


class Pif(ptt.Pattern):
    ...



"""Kernel.sc & Main.sc"""

import enum
import threading
import atexit
import pathlib
import time
import random
import sys

import sc3
from . import _taskq as tsq
from . import platform as plf
from . import systemactions as sac
from . import stream as stm
from . import _oscinterface as osci
from . import classlibrary as clb
from . import clock as clk

# Have late_imports and is not in the above tree.
from ..seq import pattern as _


__all__ = ['main', 'RtMain', 'NrtMain']


main = None
'''Default main class global variable set by sc3.init().'''


class TimeException(ValueError):
    pass


### Kernel.sc ###


class Process(type):
    RT_MODE = 0
    NRT_MODE = 1

    class _atexitprio(enum.IntEnum):
        ''' Library predefined _atexitq priority numbers.'''
        CUSTOM = 0
        SERVERS = 500
        PLATFORM = 700
        CLOCKS = 800
        NETWORKING = 900

    _atexitq = tsq.TaskQueue()
    '''Functions registered in atexit with order by priority numbers.'''

    def __init__(cls, name, bases, dict):
        # Main library lock.
        cls._main_lock = threading.RLock()

        # Mode switch lock. Not defined behaviour.
        # cls._switch_cond = threading.Condition(cls._main_lock)

        # Main TimeThread random generator.
        cls._m_rgen = random.Random()

        # SynthDef graph build's global state.
        cls._current_synthdef = None
        cls._def_build_lock = threading.Lock()

        cls._init_platform()
        atexit.register(cls._shutdown)

    def _init_platform(cls):
        if sys.platform.startswith('linux'):
            cls._platform = plf.LinuxPlatform()
        elif sys.platform.startswith('darwin'):
            cls._platform = plf.DarwinPlatform()
        elif sys.platform.startswith('win32'):
            cls._platform = plf.Win32Platform()
        elif sys.platform.startswith('cygwin'):
            cls._platform = plf.CygwinPlatform()
        else:
            raise RuntimeError('platform not defined')
        cls._platform._startup()
        cls._atexitq.add(cls._atexitprio.PLATFORM, cls.platform._shutdown)

    def _startup(cls):
        raise NotImplementedError

    def _exec_startup_file(cls):
        # NOTE: This may change to a special kind
        # of file that only allows to set up certain
        # parameters like platform or server options.
        if sc3.LIB_SETUP_FILE:
            path = pathlib.Path(sc3.LIB_SETUP_FILE)
        else:
            path = plf.Platform.config_dir / 'startup.py'
        if path.exists():
            with open(path, 'r') as file:
                ast = compile(file.read(), path, 'exec')
                exec(ast, dict(), dict())

    def _shutdown(cls):
        sac.ShutDown.run()
        while not cls._atexitq.empty():
            cls._atexitq.pop()[1]()
        atexit.unregister(cls._shutdown)

    @property
    def _rgen(cls):
        return cls.current_tt._rgen

    @property
    def platform(cls):
        return cls._platform

    def open_udp_port(cls, port):
        raise NotImplementedError('multiple UDP ports are not implemented')

    def add_osc_recv_func(cls, func):
        cls._osc_interface.add_recv_func(func)

    def remove_osc_recv_func(cls, func):
        cls._osc_interface.remove_recv_func(func)

    def elapsed_time(cls):
        pass

    def update_logical_time(cls, seconds=None):
        pass


### Main.sc ###


class RtMain(metaclass=Process):
    @classmethod
    def _init(cls):
        # In sclang these two are the same clock, it obtains time_since_epoch
        # as gHostStartNanos = duration_cast<nanoseconds>
        # (hrTimeOfInitialization.time_since_epoch()).count().
        # I think is not important for these reference points to be sampled at
        # the same time because main.elapsed_time() == 0 is in logic relation
        # to main._time_of_initialization. Comment to be removed later if
        # true.
        cls._time_of_initialization = time.time()  # time_since_epoch
        cls._perf_counter_time_of_initialization = time.perf_counter()  # monotonic clock.
        cls.main_tt = stm._MainTimeThread()
        cls.current_tt = cls.main_tt
        cls._osc_interface = osci.OscUdpInterface(
            sc3.LIB_PORT, sc3.LIB_PORT_RANGE)
        cls._osc_interface.start()

        clb.ClassLibrary.init()
        cls._startup()

    @classmethod
    def _startup(cls):
        import sc3.synth.systemdefs as sds
        sds.SystemDefs.add_all()
        cls._exec_startup_file()
        sac.StartUp.run()

    @classmethod
    def elapsed_time(cls):
        '''Physical time since library initialization.'''
        # *elapsedTime _ElapsedTime
        # Returns the more accurate clock time minus _time_of_initialization.
        return time.perf_counter() - cls._perf_counter_time_of_initialization

    @classmethod
    def update_logical_time(cls, seconds=None):
        # // When code is run from the code editor, the command line, or in
        # // response to OSC and MIDI messages, the main Thread's logical time
        # // is set to the current physical time (see Process: *elapsedTime).
        # Text above doesn't hold all true here, updates are done by clocks
        # when data is received but can't capture code before excecution,
        # wouldn't be prudent. Routines are required to get logical time.
        # // When code scheduled on a Clock is run, the main Thread's logical
        # // time is set to the time the code was scheduled for. Child Threads
        # // inherit logical time from their parents - whenever a Thread
        # // (Routine) is started or resumed, its logical time is set to that
        # // of the parent Thread.
        now = cls.elapsed_time()
        if seconds is None:
            # Logical time is set to physical time.
            cls.main_tt._m_seconds = now
        elif seconds > now:
            raise TimeException(
                "logical time can't be set to the future of physical time")
        else:
            # Logical time is set to current sched time by clocks.
            cls.main_tt._m_seconds = seconds


class NrtMain(metaclass=Process):
    @classmethod
    def _init(cls):
        cls._time_of_initialization = 0.0
        cls.main_tt = stm._MainTimeThread()
        cls.main_tt._m_seconds = 0.0
        cls.current_tt = cls.main_tt
        cls._clock_scheduler = clk.ClockScheduler()
        cls._osc_interface = osci.OscNrtInterface()
        cls._osc_interface.init()
        cls.osc_score = None

        clb.ClassLibrary.init()
        cls._startup()

    @classmethod
    def _startup(cls):
        # Server setup and boot is only for user convenience.
        import sc3.synth.server as srv
        srv.Server.default.latency = 0
        srv.Server.default.boot()  # Sets _status_watcher._has_booted = True
        cls._exec_startup_file()
        sac.StartUp.run()

    @classmethod
    def elapsed_time(cls):
        '''Physical time is main_Thread.seconds in nrt.'''
        return float(cls.main_tt.seconds)

    @classmethod
    def update_logical_time(cls, seconds=None):
        if seconds is None:
            return
        else:
            cls.main_tt._m_seconds = seconds

    @classmethod
    def reset(cls):
        '''Reset sc3 time, scheduler and osc score to initial state.'''
        cls.main_tt._m_seconds = 0.0
        cls._clock_scheduler.reset()
        cls._osc_interface.init()  # Reset OscScore.

    @classmethod
    def process(cls, tail=0):
        '''Generate and return the OSC score.'''
        cls._clock_scheduler.run()
        cls._osc_interface._osc_score.finish(tail)
        cls.osc_score = cls._osc_interface._osc_score
        return cls.osc_score



"""
Functions used to process lists as in sclang.
At the moment they are mostly for internal use.
"""

import itertools
import operator


def as_list(obj):
    '''
    Bubble non iterable objects, tuples and strings in a list. If obj is None
    returns an empty list. For the rest of iterators is the same as list(obj).
    '''
    if isinstance(obj, (tuple, str)):
        return [obj]
    elif hasattr(obj, '__iter__'):
        return list(obj)
    else:
        return [obj]


def unbubble(obj):
    '''If obj is a list of one item or any other object unbubble(obj)
    returns the item, otherwise returns the list or object unchanged.'''
    if isinstance(obj, list) and len(obj) == 1:  # only one level
        return obj[0]
    else:
        return obj

# TODO: replace all inlist by lst for convention.

def flat(inlist):
    def _(inlist, outlist):
        for item in inlist:
            if isinstance(item, list):
                _(item[:], outlist)  # TODO: Not sure if has to be a copy.
            else:
                outlist.append(item)
    outlist = []
    _(inlist, outlist)
    return outlist


def flatten(inlist, n_levels=1):
    def _(inlist, outlist, n):
        for item in inlist:
            if n < n_levels:
                if isinstance(item, list):
                    _(item[:], outlist, n + 1)  # TODO: Not sure if has to be a copy.
                else:
                    outlist.append(item)
            else:
                outlist.append(item)
    outlist = []
    _(inlist, outlist, 0)
    return outlist


def shape(lst):
    if isinstance(lst, (list, tuple)):
        size = len(lst)
        if size > 0:
            # // This assumes every element has the same shape.
            sub_shape = shape(lst[0])
            if len(sub_shape) > 0:
                ret = [size]
                ret.extend(sub_shape)
                return tuple(ret)
            else:
                return (size, )
        else:
            return (0,)
    else:
        return tuple()


def reshape_like(one, another):
    index = 0
    one_flat = flat(one)

    def func(*discard):
        nonlocal index
        item = one_flat[index % len(one_flat)]
        index += 1
        return item
    return deep_collect(another, 0x7FFFFFFF, func)


def deep_collect(inlist, depth, func, index=0, rank=0):
    if depth is None:
        rank += 1
        if isinstance(inlist, list):
            return [
                deep_collect(item, depth, func, i, rank)
                for i, item in enumerate(inlist)]
        else:
            return func(inlist, index, rank)
    if depth <= 0:
        if func:
            return func(inlist, index, rank)
        else:
            return None
    depth -= 1
    rank += 1
    if isinstance(inlist, list):
        return [
            deep_collect(item, depth, func, i, rank)
            for i, item in enumerate(inlist)]
    else:
        return func(inlist, index, rank)


def extend(lst, n, item):
    '''
    Create a new list by extending lst with item.
    '''
    l = len(lst)
    if l == 0 or n <= 0: return []
    return (lst + [item] * (n - l))[:n]


def wrap_extend(lst, n):
    '''
    Create a new list by extending lst with its own elements cyclically.
    '''
    l = len(lst)
    if l == 0 or n <= 0: return []
    return lst * (n // l) + lst[:n % l]


def list_unop(op, a, t=None):
    t = t or list
    t_seq = (list, tuple)  # TODO: check.
    if isinstance(a, t_seq):
        if any(isinstance(i, t_seq) for i in a):
            return t(list_unop(op, i, type(i)) for i in a)
        return t(op(i) for i in a)
    else:
        return op(a)


def list_binop(op, a, b, t=None):
    '''
    Operate on sequences element by element and return a sequence of type t
    (default to list) if any argument a or b is sequence. Arguments a and b may
    be types or a sequence of types compatible with operator op. Nested
    sequences are processed recursively. Recursion uses tuple type cast if any
    is tuple or type(a) if is a is sequence else type(b) if b is sequence.
    Tuples are termporary converted to lists to process. All this is needed
    because multichannel expansion behaviour (e.g. in Env).
    '''
    # NOTE: The other option is to leave tuples as tuples, could be
    # better, the problem is outside UGen operations as in Env.
    t = t or list
    t_seq = (list, tuple)  # TODO: check.
    if isinstance(a, t_seq) and isinstance(b, t_seq):
        if len(a) >= len(b):
            b = wrap_extend(list(b), len(a))
        else:
            a = wrap_extend(list(a), len(b))
        if any(isinstance(i, t_seq) for i in a)\
        or any(isinstance(i, t_seq) for i in b):
            ret = []
            a2 = b2 = t2 = None
            for i in range(len(a)):
                if isinstance(a[i], tuple):
                    t2 = tuple
                    a2 = list(a[i])
                if isinstance(b[i], tuple):
                    t2 = tuple
                    b2 = list(b[i])
                if t2 is None:
                    if isinstance(a[i], t_seq):
                        t2 = list
                    elif isinstance(b[i], t_seq):
                        t2 = list
                a2 = a2 or a[i]
                b2 = b2 or b[i]
                t2 = t2 or type(...)  # if neither is t_seq type doesn't matter but can't be None.
                ret.append(list_binop(op, a2, b2, t2))
                a2 = b2 = t2 = None
            return t(ret)
        else:
            return t(op(i[0], i[1]) for i in zip(a, b))
    elif isinstance(a, t_seq):
        return t(list_binop(op, item_a, b, type(item_a)) for item_a in a)
    elif isinstance(b, t_seq):
        return t(list_binop(op, a, item_b, type(item_b)) for item_b in b)
    else:
        return op(a, b)


def list_narop(op, a, *args, t=None):
    t = t or list
    t_seq = (list, tuple)  # TODO: check.
    if isinstance(a, t_seq):
        if any(isinstance(i, t_seq) for i in a):
            return t(list_narop(op, i, *args, t=type(i)) for i in a)
        return t(op(i, *args) for i in a)
    else:
        return op(a, *args)


def list_sum(lst, t=None):
    res = 0
    for item in lst:
        res = list_binop(operator.add, res, item, t)
    return res

# NOTE: maxItem used in Env.duration gives the clue that only one level of
# nesting is supported by multichannels graphs. Same happens when building
# controls in SynthDef.
# [1, 2, 3, 4].maxItem // ok
# 3 > [3, 4] // ok
# [3, 4] > 3 // ok
# [1, 2, [3, 4]].maxItem // not ok
# [[1, 2], [3, 4]].maxItem // not ok

def list_min(lst, t=None):
    t_seq = (list, tuple)
    min_item = lst[0]
    if isinstance(min_item, t_seq):
        min_item = list_min(min_item, t)
    for item in lst[1:]:
        if isinstance(item, t_seq):
            item = list_min(item, t)
        if list_binop(operator.lt, item, min_item, t):
            min_item = item
    return min_item


def list_max(lst, t=None):
    t_seq = (list, tuple)
    max_item = lst[0]
    if isinstance(max_item, t_seq):
        max_item = list_max(max_item, t)
    for item in lst[1:]:
        if isinstance(item, t_seq):
            item = list_max(item, t)
        if list_binop(operator.gt, item, max_item, t):
            max_item = item
    return max_item


#def reshape_like(this, that); # or that this like sclang?

# flop [(a[i], b[i]) for i in range(len(a))] but len(a) >= len(b)
# filter() y itertools.filterfalse() are select/reject.
# select [x for x in self.control_names if x.rate == 'noncontrol']
# reject [x for x in self.control_names if x.rate != 'noncontrol']
# any with predicate (a generator) como en sclang any(x > 10 for x in l)
# Python's all()

# detect could be:
# for item in sequence:
#     if item == something:
#         return item

# clump [l[i:i + n] for i in range(0, len(l), n)] https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
# ver también grouper en Itertools Recipes
# def gen_clumps(l, n=1):
#    return (l[i:i + n] for i in range(0, len(l), n))


def clump(lst, n):
    return [lst[i:i + n] for i in range(0, len(lst), n)]


# para pairsDo:
def gen_cclumps(l, n=1):
    '''return n items as pairsDo does (with n=2) for iteration, cclump stands
    for complete clump, it discards possible non full clump at the end'''
    return (l[i:i + n] for i in range(0, len(l) - (n - 1), n))


# doAdjacentPairs, Itertools Recipes: https://docs.python.org/3/library/itertools.html#itertools-recipes
def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


# dup [UGen] * n
# collect if else ['todo' if isinstance(x, A) else 'nada' for x in arr]
# [0] * (len(names) - len(rates)) # VER: sclang extend, pero no trunca
# [l[i%len(l)] for i in range(8)] # wrapExtend

# Cartesian product, is combinatorial (itertools.product).
# https://stackoverflow.com/questions/10973766/understanding-the-map-function
# [(a, b) for a in iterable_a for b in iterable_b]

# map as collect
# map(función, iterable)
# [f(x) for x in iterable]


def lace(lst, size=None):
    if size is None:
        return [x for sub in zip(*lst) for x in sub]
    else:
        gen = itertools.cycle(x for sub in zip(*lst) for x in sub)
        return [next(gen) for _ in range(0, size)]


def flop(lst):
    lst = [[None] if x is None else as_list(x) for x in lst]  # NOTE: as_list converts None in []
    n = len(lst)
    if n == 0:
        return [[]]  # NOTE: empty column as in sclang.
    length = max(len(l) for l in lst)
    ret = [[None for _ in range(n)] for _ in range(length)]
    for i in range(length):
        for j in range(n):
            try:
                ret[i][j] = lst[j][i % len(lst[j])]  # NOTE: i % 0 == 0 in sclang, and is also empty list.
            except ZeroDivisionError:
                ret[i][j] = []  # NOTE: a = []; a[0] is nil, 0 or nil is [].
    return ret


# # other flop.
# # [[None] * length] * n
# def flop(lst):
#     # agregar lst = [as_list(x) for x in lst]
#     n = len(lst)
#     length = max(len(l) for l in lst)
#     aux = []
#     ret = []
#     for i in range(length):
#         for j in range(n):
#             aux.append(lst[j][i % len(lst[j])])
#         ret.append(aux)
#         aux = []
#     return ret
# # lst = [[1, 2], [10, 20, 30], [100, 200, 300, 400]]
# # [[lst[j][i % len(lst[j])] for j in range(len(lst))] for i in range(max(len(l) for l in lst))]


def flop_together(*lsts):  # BUG: in sclang, modifies the original list with garbage.
    max_size = 0
    for sub_list in lsts:
        for each in sub_list:
            if isinstance(each, list):
                max_size = max(max_size, len(each))
    stand_in = [None] * max_size
    ret = []
    for i, sub_list in enumerate(lsts):
        sub_list = sub_list[:]
        sub_list.append(stand_in)
        ret.append([])
        for each in flop(sub_list):
            each.pop()  # // remove stand-in
            ret[i].append(each)
    return ret


def max_depth(lst):
    def find_max(lst, max_rank=1):
        ret = max_rank
        for item in lst:
            if isinstance(item, list):
                ret = max(ret, find_max(item, max_rank + 1))
        return ret
    return find_max(lst)


def max_size_at_depth(lst, rank):
    if rank <= 0:  # BUG: in sclang recursive condition is wrong.
        return len(lst)
    max_size = 0
    for item in lst:
        if isinstance(item, list):
            sz = max_size_at_depth(item, rank - 1)
        else:
            sz = 1
        if sz > max_size:
            max_size = sz
    return max_size


def wrap_at_depth(lst, rank, index):
    if rank <= 0:  # BUG: in sclang recursive condition is wrong.
        return lst[index % len(lst)]
    ret = []
    for item in lst:  # BUG: in sclang i is not used
        if isinstance(item, list):
            ret.append(wrap_at_depth(item, rank - 1, index))
        else:
            ret.append(item)
    return ret


def flop_deep(lst, rank):
    if rank is None:
        rank = max_depth(lst) - 1
    if rank <= 1:
        return flop(lst)
    size = len(lst)
    max_size = max_size_at_depth(lst, rank)
    ret = []
    for i in range(max_size):
        ret.append(wrap_at_depth(lst, rank, i))
    return ret


def multichannel_expand_tuple(tpl, rank):
    # // Allow to multichannel expand ugen specs, like those of Klank,
    # // in the case of which two is the rank, but could be otherwise.
    lst = list(tpl)
    if max_size_at_depth(lst, rank) <= 1:
        return tpl
    lst = flop_deep(lst, rank)
    return unbubble([tuple(item) for item in lst])  # why unbubble?



"""Clock.sc"""

import logging
import threading
import time as _time
import sys
import traceback
import weakref

from . import classlibrary as clb
from . import main as _libsc3
from . import builtins as bi
from . import functions as fn
from . import systemactions as sac
from . import model as mdl
from . import _taskq as tsq
from . import stream as stm


__all__ = [
    'SystemClock', 'Scheduler', 'AppClock', 'Quant', 'TempoClock', 'defer']


_logger = logging.getLogger(__name__)


class ClockError(RuntimeError):
    pass


class ClockNotRunning(ClockError):
    pass


### Clocks for timing threads ###


class MetaClock(type):
    _pure_nrt = False  # Must be set by sub metaclasses in __init__.

    def play(cls, task, quant=None):
        # Unused quant (needed for tempo clock compatibility).
        cls.sched(0, task)

    @property
    def mode(cls):
        if cls._pure_nrt:
            return _libsc3.main.NRT_MODE
        elif _libsc3.main is _libsc3.RtMain:
            return _libsc3.main.RT_MODE
        else:
            return _libsc3.main.NRT_MODE

    @property
    def seconds(cls):
        return _libsc3.main.current_tt.seconds

    # // tempo clock compatibility

    @property
    def beats(cls):
        return _libsc3.main.current_tt.seconds

    def beats2secs(cls, beats):
        return beats

    def secs2beats(cls, secs):
        return secs

    def beats2bars(cls):
        return 0

    def bars2beats(cls):
        return 0

    def time_to_next_beat(cls, quant=1):
        return 0

    def next_time_on_grid(cls, quant=1, phase=0):
        if quant == 0:
            return cls.beats() + phase
        if phase < 0:
            phase = bi.mod(phase, quant)
        return bi.roundup(cls.beats() - bi.mod(phase, quant), quant) + phase


class Clock(metaclass=MetaClock):
    pass


class MetaSystemClock(MetaClock):
    def __init__(cls, *_):

        def init_func(cls):
            if _libsc3.main is _libsc3.RtMain:
                cls._task_queue = tsq.TaskQueue()
                cls._sched_cond = threading.Condition(_libsc3.main._main_lock)
                cls._thread = threading.Thread(
                    target=cls._run,
                    name=cls.__name__,
                    daemon=True)
                cls._thread.start()
                cls._sched_init()
                _libsc3.main._atexitq.add(
                    _libsc3.main._atexitprio.CLOCKS, cls._sched_stop)
            else:
                cls._pure_nrt = True
                cls._elapsed_osc_offset = 0.0

        clb.ClassLibrary.add(cls, init_func)


class SystemClock(Clock, metaclass=MetaSystemClock):
    _SECONDS_FROM_1900_TO_1970 = 2208988800 # (int32)UL # 17 leap years
    _NANOS_TO_OSC = 4.294967296 # PyrSched.h: const double kNanosToOSC  = 4.294967296; // pow(2,32)/1e9
    _MICROS_TO_OSC = 4294.967296 # PyrSched.h: const double kMicrosToOSC = 4294.967296; // pow(2,32)/1e6
    _SECONDS_TO_OSC = 4294967296. # PyrSched.h: const double kSecondsToOSC  = 4294967296.; // pow(2,32)/1
    _OSC_TO_NANOS = 0.2328306436538696# PyrSched.h: const double kOSCtoNanos  = 0.2328306436538696; // 1e9/pow(2,32)
    _OSC_TO_SECONDS = 2.328306436538696e-10 # PyrSched.h: const double kOSCtoSecs = 2.328306436538696e-10;  // 1/pow(2,32)

    def __new__(cls):
        return cls

    @classmethod
    def _sched_init(cls):
        # _time_of_initialization was moved to main because rt/nrt clock switch.
        cls._host_osc_offset = 0 # int64
        cls._sync_osc_offset_with_tod()
        cls._host_start_nanos = int(_libsc3.main._time_of_initialization / 1e-9)
        cls._elapsed_osc_offset = int(
            cls._host_start_nanos * cls._NANOS_TO_OSC
        ) + cls._host_osc_offset

        # same every 20 secs
        cls._resync_cond = threading.Condition()
        cls._resync_thread = threading.Thread(
            target=cls._resync_thread_func,
            name=f'{cls.__name__}.resync',
            daemon=True)
        cls._resync_thread.start()

    @classmethod
    def _sync_osc_offset_with_tod(cls):
        # // generate a value gHostOSCoffset such that
        # // (gHostOSCoffset + systemTimeInOSCunits)
        # // is equal to gettimeofday time in OSCunits.
        # // Then if this machine is synced via NTP
        # // we are synced with the world.
        # // more accurate way to do this??
        number_of_tries = 8
        diff = 0 # int64
        min_diff = 0x7fffFFFFffffFFFF # int64
        new_offset = cls._host_osc_offset

        for i in range(0, number_of_tries):
            system_time_before = _time.time()  # must be the same epoch
            time_of_day = _time.time()  # must be gmt time
            system_time_after = _time.time()

            system_time_before = int(system_time_before / 1e-9)
            system_time_after = int(system_time_after / 1e-9)
            diff = system_time_after - system_time_before

            if diff < min_diff:
                min_diff = diff

                system_time_between = system_time_before + diff // 2
                system_time_in_osc_units = int(
                    system_time_between * cls._NANOS_TO_OSC)

                # mimics
                tv_sec = int(time_of_day)
                tv_usec = int((time_of_day % 1) / 1e-6)
                time_of_day_in_osc_units = (
                    (int(tv_sec + cls._SECONDS_FROM_1900_TO_1970) << 32)
                    + int(tv_usec * cls._MICROS_TO_OSC))

                new_offset = time_of_day_in_osc_units - system_time_in_osc_units

        # This function seems to add only the jitter between clock calls here
        # and in sclang. Running every 20 seconds or 1 minute would be the same
        # if the date clock wasn't adjusted by ntp. Based on the logged numbers
        # osc time may jump up to ~700 nanoseconds, far from ~238 picoseconds.
        # NOTE: Without this bundles will be late.
        # print('new offset diff:', cls._host_osc_offset - new_offset)
        cls._host_osc_offset = new_offset

    @classmethod
    def _resync_thread_func(cls):  # L408
        cls._run_resync = True

        with cls._resync_cond:
            while cls._run_resync:
                cls._resync_cond.wait(20)
                if not cls._run_resync:
                    return

                cls._sync_osc_offset_with_tod()
                cls._elapsed_osc_offset = int(
                    cls._host_start_nanos * cls._NANOS_TO_OSC
                ) + cls._host_osc_offset

    @classmethod
    def elapsed_time_to_osc(cls, elapsed: float) -> int:  # int64
        return int(
            elapsed * cls._SECONDS_TO_OSC
        ) + cls._elapsed_osc_offset

    @classmethod
    def osc_to_elapsed_time(cls, osctime: int) -> float:
        return float(
            osctime - cls._elapsed_osc_offset
        ) * cls._OSC_TO_SECONDS

    @classmethod
    def osc_time(cls) -> int:
        return cls.elapsed_time_to_osc(_libsc3.main.elapsed_time())

    @classmethod
    def _sched_add(cls, secs, task):
        # Call with acquired lock.
        if cls._task_queue.empty():
            prev_time = -1e10
        else:
            prev_time = cls._task_queue.peek()[0]
        cls._task_queue.add(secs, task)
        # if isinstance(task, pst.PauseStream):
        #     task._next_beat = secs
        if cls._task_queue.peek()[0] != prev_time:
            cls._sched_cond.notify_all()

    @classmethod
    def _sched_stop(cls):
        if not cls._run_sched:
            return
        with cls._sched_cond:
            with cls._resync_cond:
                if cls._run_resync:
                    cls._run_resync = False
                    cls._resync_cond.notify()
            cls._task_queue.clear()
            cls._run_sched = False
            cls._sched_cond.notify_all()
        cls._resync_thread.join()
        cls._thread.join()

    @classmethod
    def _run(cls):
        cls._run_sched = True

        with cls._sched_cond:
            while True:
                # // wait until there is something in scheduler
                while cls._task_queue.empty():
                    cls._sched_cond.wait()
                    if not cls._run_sched:
                        return

                # // wait until an event is ready
                now = 0
                while not cls._task_queue.empty():
                    # NOTE: I think there is no need for this clock to be
                    # monotonic and sclang may or may not be working this way,
                    # is less complicated here to use just elapsed_time.
                    # I leave previous code commented for now (this should be the same).
                    # now = _time.time()
                    now = _libsc3.main.elapsed_time()
                    sched_secs = cls._task_queue.peek()[0]
                    # sched_point = (_libsc3.main._time_of_initialization
                    #                + sched_secs)
                    # if now >= sched_point:
                    if now >= sched_secs:
                        break
                    # cls._sched_cond.wait(sched_point - now)
                    cls._sched_cond.wait(sched_secs - now)
                    if not cls._run_sched:
                        return

                # // perform all events that are ready
                # while not cls._task_queue.empty()\
                # and now >= (_libsc3.main._time_of_initialization
                #             + cls._task_queue.peek()[0]):
                while not cls._task_queue.empty()\
                and now >= cls._task_queue.peek()[0]:
                    item = cls._task_queue.pop()
                    sched_time = item[0]
                    task = item[1]
                    # if isinstance(task, pst.PauseStream):
                    #     task._next_beat = None
                    try:
                        _libsc3.main.update_logical_time(sched_time)
                        delta = task.__awake__(sched_time, sched_time, cls)
                        if isinstance(delta, (int, float))\
                        and not isinstance(delta, bool):
                            time = sched_time + delta
                            cls._sched_add(time, task)
                    except stm.StopStream:
                        pass
                    except Exception:
                        # Always recover.
                        _logger.error(
                            '%s(%s) scheduled on SystemClock',
                            type(task).__name__, task.func.__qualname__,
                            exc_info=1)

    @classmethod
    def clear(cls):
        if cls.mode == _libsc3.main.NRT_MODE:
            return
        with cls._sched_cond:
            while not cls._task_queue.empty():
                cls._task_queue.pop()
            cls._sched_cond.notify_all()

    @classmethod
    def sched(cls, delta, item):
        if not hasattr(item, '__awake__'):
            item = fn.Function(item)
        item._clock = cls
        if cls.mode == _libsc3.main.NRT_MODE:
            seconds = _libsc3.main.current_tt.seconds
            seconds += delta
            if seconds == float('inf'):
                return
            ClockTask(seconds, cls, item, _libsc3.main._clock_scheduler)
        else:
            with cls._sched_cond:
                seconds = _libsc3.main.current_tt.seconds
                seconds += delta
                if seconds == float('inf'):
                    return
                cls._sched_add(seconds, item)

    @classmethod
    def sched_abs(cls, time, item):
        if not hasattr(item, '__awake__'):
            item = fn.Function(item)
        item._clock = cls
        if time == float('inf'):
            return
        if cls.mode == _libsc3.main.NRT_MODE:
            ClockTask(time, cls, item, _libsc3.main._clock_scheduler)
        else:
            with cls._sched_cond:
                cls._sched_add(time, item)


class Scheduler():
    def __init__(self, clock, drift=False, recursive=True):
        self._clock = clock
        self._drift = drift
        self.recursive = recursive
        # init
        self._beats = 0.0
        self._seconds = 0.0
        self.queue = tsq.TaskQueue()
        self._expired = []

    def _wakeup(self, item):
        try:
            delta = item.__awake__(self._beats, self._seconds, self._clock)
            if isinstance(delta, (int, float)) and not isinstance(delta, bool):
                self._sched_add(delta, item)
        except stm.StopStream:
            pass
        except Exception:
            _logger.error(
                '%s(%s) scheduled on AppClock',
                type(item).__name__, item.func.__qualname__, exc_info=1)

    def play(self, task):
        self.sched(0, task)

    def _sched_add(self, delta, item):
        if self._drift:
            from_time = _libsc3.main.elapsed_time()
        else:
            from_time = self.seconds = _libsc3.main.current_tt.seconds
        self.queue.add(from_time + delta, item)

    def sched(self, delta, item):
        if not hasattr(item, '__awake__'):
            item = fn.Function(item)
        item._clock = self._clock
        if delta is None:
            delta = 0.0
        if delta == float('inf'):
            return
        self._sched_add(delta, item)

    def sched_abs(self, time, item):
        if not hasattr(item, '__awake__'):
            item = fn.Function(item)
        item._clock = self._clock
        if time == float('inf'):
            return
        self.queue.add(time, item)

    def clear(self):
        while not self.queue.empty():
            self.queue.pop()

    def empty(self):
        return self.queue.empty()

    def advance(self, delta):
        self.seconds = self.seconds + delta

    @property
    def seconds(self):
        return self._seconds

    @seconds.setter
    def seconds(self, value):
        if self.queue.empty():
            self._seconds = value
            self._beats = self._clock.secs2beats(value)
            return
        self._seconds = self.queue.peek()[0]
        if self.recursive:
            while self._seconds <= value:
                self._beats = self._clock.secs2beats(self._seconds)
                self._wakeup(self.queue.pop()[1])
                if self.queue.empty():
                    break
                else:
                    self._seconds = self.queue.peek()[0]
        else:
            # // First pop all the expired items and only then wake
            # // them up, in order for control to return to the caller
            # // before any tasks scheduled as a result of this call are
            # // awaken.
            while self._seconds <= value:
                self._expired.append(self.queue.pop())
                if self.queue.empty():
                    break
                else:
                    self._seconds = self.queue.peek()[0]
            for time, item in self._expired:
                self._seconds = time
                self._beats = self._clock.secs2beats(time)
                self._wakeup(item)
            self._expired.clear()
        self._seconds = value
        self._beats = self._clock.secs2beats(value)


class MetaAppClock(MetaClock):
    def __init__(cls, *_):

        def init_func(cls):
            if _libsc3.main is _libsc3.RtMain:
                cls._sched_cond = threading.Condition(_libsc3.main._main_lock)
                cls._tick_cond = threading.Condition()
                cls._scheduler = Scheduler(cls, drift=True, recursive=False)
                cls._thread = threading.Thread(
                    target=cls._run,
                    name=cls.__name__,
                    daemon=True)
                cls._thread.start()
                _libsc3.main._atexitq.add(
                    _libsc3.main._atexitprio.CLOCKS + 1, cls._stop)
            else:
                cls._pure_nrt = True

        clb.ClassLibrary.add(cls, init_func)


class AppClock(Clock, metaclass=MetaAppClock):
    def __new__(cls):
        return cls

    @classmethod
    def _run(cls):
        cls._run_sched = True
        seconds = None
        while True:
            with cls._sched_cond:
                seconds = cls._tick()  # First tick for free, returns None
                if isinstance(seconds, (int, float))\
                and not isinstance(seconds, bool):
                    seconds = seconds - cls._scheduler.seconds  # tick returns abstime (elapsed)
            with cls._tick_cond:  # many notify one wait
                cls._tick_cond.wait(seconds)  # if seconds is None waits for notify
            if not cls._run_sched: return

    @classmethod
    def clear(cls):
        if cls.mode == _libsc3.main.NRT_MODE:
            return
        else:
            cls._scheduler.clear()

    @classmethod
    def sched(cls, delta, item):
        if cls.mode == _libsc3.main.NRT_MODE:
            if not hasattr(item, '__awake__'):
                item = fn.Function(item)
            item._clock = cls
            if delta == float('inf'):
                return
            ClockTask(delta, cls, item, _libsc3.main._clock_scheduler)
        else:
            with cls._sched_cond:
                cls._scheduler.sched(delta, item)
            with cls._tick_cond:
                cls._tick_cond.notify()

    @classmethod
    def _tick(cls):
        if cls.mode == _libsc3.main.NRT_MODE:
            return None
        cls._scheduler.seconds = _libsc3.main.elapsed_time()
        if cls._scheduler.queue.empty():
            return None  # check value for client::tick to stop calling itself.
        else:
            return cls._scheduler.queue.peek()[0]

    @classmethod
    def _stop(cls):
        if not cls._run_sched:
            return
        with cls._tick_cond:
            cls._run_sched = False
            cls._tick_cond.notify()
        cls._thread.join()


class ClockScheduler():
    def __init__(self):
        self.queue = tsq.TaskQueue()

    def run(self):
        while not self.queue.empty():
            time, clock_task = self.queue.pop()
            clock_task._wakeup(time)

    def add(self, time, clock_task):
        self.queue.add(time, clock_task)

    def reset(self):
        self.queue.clear()


class ClockTask():
    def __init__(self, beats, clock, task, scheduler):
        self.clock = clock
        self.task = task
        self.scheduler = scheduler
        scheduler.add(clock.beats2secs(beats), self)

    def _wakeup(self, time):
        try:
            # if isinstance(self.task, pst.PauseStream):
            #     self.task._next_beat = None
            _libsc3.main.update_logical_time(time)
            beats = self.clock.secs2beats(time)
            delta = self.task.__awake__(beats, time, self.clock)
            if isinstance(delta, (int, float)) and not isinstance(delta, bool):
                self.scheduler.add(self.clock.beats2secs(beats + delta), self)
        except stm.StopStream:
            pass
        except Exception:
            _logger.error(
                '%s(%s) scheduled on ClockScheduler',
                type(self.task).__name__, self.task.func.__qualname__,
                exc_info=1)


### Quant.sc ###
# // This class is used to encapsulate quantization issues associated
# // with EventStreamPlayer and TempoClock, quant and phase determine the
# // starting time of something scheduled by a TempoClock timingOffset
# // is an additional timing factor that allows an EventStream to compute
# // "ahead of time" enough to allow negative lags for strumming a chord, etc.


class Quant():
    def __init__(self, quant=0, phase=None, timing_offset=None):
        self.quant = quant
        self.phase = phase
        self.timing_offset = timing_offset

    # *default  # NOTE: Not really needed.

    @classmethod
    def as_quant(cls, quant):
        if isinstance(quant, cls):
            pass
        elif isinstance(quant, int):
            quant = cls(quant)
        elif isinstance(quant, (list, tuple)):
            quant = cls(*quant[:3])
        elif quant is None:
            quant = cls()
        else:
            msg = f'unsuported type conversion to Quant: {type(quant)}'
            raise TypeError(msg)
        return quant

    def next_time_on_grid(self, clock):
        return clock.next_time_on_grid(
            self.quant, (self.phase or 0) - (self.timing_offset or 0))

    def __repr__(self):
        return (
            f'{type(self).__name__}(quant={self.quant}, '
            f'phase={self.phase}, timing_offset={self.timing_offset})')


# /*
# You should only change the tempo 'now'. You can't set the tempo at some beat
# in the future or past, even though you might think so from the methods.
#
# There are several ideas of now:
#     elapsed time, i.e. "real time"
#     logical time in the current time base.
#     logical time in another time base.
#
# Logical time is time that is incremented by exact amounts from the time you
# started. It is not affected by the actual time your task gets scheduled, which
# may shift around somewhat due to system load. By calculating using logical
# time instead of actual time, your process will not drift out of sync over long
# periods. Every thread stores a clock and its current logical time in seconds
# and beats relative to that clock.
#
# Elapsed time is whatever the system clock says it is right now. Elapsed time
# is always advancing. Logical time only advances when your task yields or
# returns.
# */


class MetaTempoClock(MetaClock):
    def __init__(cls, *_):
        cls._all = weakref.WeakSet()

        def init_func(cls):
            # TempoClock.default was removed. Within the library use
            # SystemClock or AppClock, in user code create an instance
            # and use it as default for your script/conding session.
            # Routines set SystemClock as default.
            # cls.default = cls()
            # cls.default.permanent = True
            sac.CmdPeriod.add(cls.__on_cmd_period)

        clb.ClassLibrary.add(cls, init_func)

    def __on_cmd_period(cls):
        for clock in cls.all:
            clock.clear()
            if not clock.permanent:
                clock.stop()

    @property
    def all(cls):
        return set(cls._all)

    def stop_all(cls):
        for clock in list(cls._all):
            clock.stop()


class TempoClock(Clock, metaclass=MetaTempoClock):
    """Tempo based scheduler.

    TempoClock is a scheduler like SystemClock, but it schedules
    relative to a tempo in beats per second.

    Parameters
    ----------
    tempo : int | float
        The initial tempo. Defaults to 1.
    beats : int | float
        The time in beats, corresponding to the reference time given
        with the seconds argument. Defaults to 0.
    seconds : int | float
        The reference time in seconds, to which the beats argument
        corresponds. Defaults to the current Thread's logical time.

    Notes
    -----
    The TempoClock will be created as if it started counting beats at
    the time given in the seconds argument with the starting amount
    givenin the beats argument. The current count of beats will thus
    be equal to that starting amount plus the amount of beats that
    would be counted since the given reference time in seconds,
    according to the given tempo.

    The default arguments create a TempoClock that starts counting
    beats with 0 at the current logical time.
    ::

      @routine.run()
      def example():
          # Starts from zero by default.
          t = TempoClock(1)
          print('current beats:', t.beats)
          # Starts counting beats from 5.
          t = TempoClock(1, 5)
          print('current beats:', t.beats)
          # Counting beats as if it started 5 seconds ago from 0.
          t = TempoClock(1, 0, main.current_tt.seconds - 5)
          print('current beats:', t.beats)

    If the above example was run without a souronding routine the
    actual `beats` value will not be precise. When running in real
    time, the base time reference is elapsed time and every call that
    requires current time is relativelly synced to it. The same
    happens in sclang when each instruction is evaluated separatelly,
    the difference is that sclang updates the base time only once per
    intepreter call which is not possible to do in a python library.
    """

    def __init__(self, tempo=None, beats=None, seconds=None):
        # prTempoClock_New
        tempo = tempo or 1.0  # tempo=0 is invalid too.
        if tempo < 0.0:
            raise ValueError(f'invalid tempo {tempo}')

        self._tempo = tempo
        self._beat_dur = 1.0 / tempo
        self._base_seconds = seconds or _libsc3.main.current_tt.seconds
        self._base_beats = beats or 0.0
        self._beats = 0.0

        # init
        self._beats_per_bar = 4.0
        self._bars_per_beat = 0.25
        self._base_bar_beat = 0
        self._base_bar = 0.0
        self.permanent = False
        type(self)._all.add(self)

        if _libsc3.main is _libsc3.RtMain:
            self._pure_nrt = False
            self._task_queue = tsq.TaskQueue()
            self._sched_cond = threading.Condition(_libsc3.main._main_lock)
            self._thread = threading.Thread(
                target=self._run,
                name=f'{type(self).__name__} id: {id(self)}',
                daemon=True)
            self._thread.start()
            _libsc3.main._atexitq.add(
                _libsc3.main._atexitprio.CLOCKS + 2, self._stop)
        else:
            self._pure_nrt = True

    @property
    def mode(self):
        if self._pure_nrt:
            return _libsc3.main.NRT_MODE
        elif _libsc3.main is _libsc3.RtMain:
            return _libsc3.main.RT_MODE
        else:
            return _libsc3.main.NRT_MODE

    def _run(self):
        self._run_sched = True

        with self._sched_cond:
            while True:
                # // wait until there is something in scheduler
                while self._task_queue.empty():
                    self._sched_cond.wait()
                    if not self._run_sched:
                        return

                # // wait until an event is ready
                elapsed_beats = 0
                while not self._task_queue.empty():
                    elapsed_beats = self.elapsed_beats()
                    qpeek = self._task_queue.peek()
                    if elapsed_beats >= qpeek[0]:
                        break
                    sched_secs = self.beats2secs(qpeek[0])
                    # NOTE: I think there is no need for this clock to be
                    # monotonic and sclang may or may not be working this way,
                    # is less complicated here to use just elapsed_time.
                    # I leave previous code commented by now (this should be
                    # the same).
                    # sched_point = _libsc3.main._time_of_initialization + sched_secs
                    # self._sched_cond.wait(sched_point - _time.time())
                    self._sched_cond.wait(
                        sched_secs - _libsc3.main.elapsed_time())
                    if not self._run_sched:
                        return

                # // perform all events that are ready
                while not self._task_queue.empty()\
                and elapsed_beats >= self._task_queue.peek()[0]:
                    item = self._task_queue.pop()
                    self._beats = item[0]
                    task = item[1]
                    # if isinstance(task, pst.PauseStream):
                    #     task._next_beat = None
                    try:
                        _libsc3.main.update_logical_time(
                            self.beats2secs(self._beats))
                        delta = task.__awake__(
                            self._beats, self.beats2secs(self._beats), self)
                        if isinstance(delta, (int, float))\
                        and not isinstance(delta, bool):
                            time = self._beats + delta
                            self._sched_add(time, task)
                    except stm.StopStream:
                        pass
                    except Exception:
                        _logger.error(
                            '%s(%s) scheduled on TempoClock id %s',
                            type(task).__name__, task.func.__qualname__,
                            id(self), exc_info=1)

    def stop(self):
        if self.mode == _libsc3.main.NRT_MODE:
            return
        if not self.running():
            _logger.debug(f'{self} is not running')
            return
        stop_thread = threading.Thread(
            target=self._stop,
            name=f'{type(self).__name__}.stop_thread id: {id(self)}',
            daemon=True)
        stop_thread.start()
        _libsc3.main._atexitq.remove(self._stop)

    def _stop(self):
        if not self._run_sched:
            return
        with self._sched_cond:
            self._task_queue.clear()
            type(self)._all.remove(self)
            self._run_sched = False
            self._sched_cond.notify_all()
        self._thread.join()
        self._thread = None
        self._sched_cond = None

    # def __del__(self):
    #     # BUG: self needs to be stoped to be gc collected because is in
    #     # BUG: cyclic reference with target=_run.
    #     self.stop()

    def play(self, task, quant=None):
        quant = Quant.as_quant(quant)
        self.sched_abs(quant.next_time_on_grid(self), task)

    def play_next_bar(self, task):
        self.sched_abs(self.next_bar(), task)

    @property
    def tempo(self):
        # _TempoClock_Tempo
        if not self.running():
            raise ClockNotRunning(self)
        return self._tempo

    @tempo.setter
    def tempo(self, value):
        # // For setting the tempo at the current logical time
        # // (even another TempoClock's logical time).
        # setTempoAtBeat(newTempo, this.beats) -> prTempoClock_SetTempoAtBeat
        if not self.running():
            raise ClockNotRunning(self)
        if value == 0.0:
            raise ValueError("tempo can't be zero")
        if self._tempo < 0.0:
            raise ValueError(
                "cannot set tempo from this method. "
                "The method 'etempo()' can be used instead")
        if value < 0.0:
            raise ValueError(
                f"invalid tempo {value}. The method "
                "'etempo()' can be used instead.")
        # TempoClock::SetTempoAtBeat
        beats = self.beats
        self._base_seconds = self.beats2secs(beats)
        self._base_beats = beats
        self._tempo = value
        self._beat_dur = 1.0 / self._tempo
        # en tempo_
        mdl.NotificationCenter.notify(self, 'tempo')
        if self.mode == _libsc3.main.NRT_MODE:
            return
        else:
            with self._sched_cond:
                self._sched_cond.notify()  # NOTE: is notify_one in C++.

    def etempo(self, value):
        # // For setting the tempo at the current elapsed time.
        # TODO: this.setTempoAtSec(newTempo, Main.elapsedTime);
        # _TempoClock_SetTempoAtTime
        if not self.running():
            raise ClockNotRunning(self)
        if value == 0.0:
            raise ValueError("tempo can't be zero")
        # TempoClock::SetTempoAtTime
        seconds = _libsc3.main.elapsed_time()
        self._base_beats = self.secs2beats(seconds)
        self._base_seconds = seconds
        self._tempo = value
        self._beat_dur = 1.0 / self._tempo
        # etempo_
        mdl.NotificationCenter.notify(self, 'tempo')
        if self.mode == _libsc3.main.NRT_MODE:
            return
        else:
            with self._sched_cond:
                self._sched_cond.notify()  # NOTE: is notify_one in C++.

    def beat_dur(self):
        if not self.running():
            raise ClockNotRunning(self)
        return self._beat_dur

    def elapsed_beats(self):
        if not self.running():
            raise ClockNotRunning(self)
        return self.secs2beats(_libsc3.main.elapsed_time())

    @property
    def beats(self):
        """Returns the appropriate beats for this clock from any thread."""
        return self.secs2beats(_libsc3.main.current_tt.seconds)

    @beats.setter
    def beats(self, value):
        if not self.running():
            raise ClockNotRunning(self)
        seconds = _libsc3.main.current_tt.seconds
        self._base_seconds = seconds
        self._base_beats = value
        self._beat_dur = 1.0 / self._tempo
        if self.mode == _libsc3.main.NRT_MODE:
            return
        else:
            with self._sched_cond:
                self._sched_cond.notify()  # NOTE: is notify_one in C++

    @property
    def seconds(self):
        return _libsc3.main.current_tt.seconds

    def _sched_add(self, beats, task):
        # Call with acquired lock.
        if self._task_queue.empty():
            prev_beat = -1e10
        else:
            prev_beat = self._task_queue.peek()[0]
        self._task_queue.add(beats, task)
        # if isinstance(task, pst.PauseStream):
        #     task._next_beat = beats
        if self._task_queue.peek()[0] != prev_beat:
            self._sched_cond.notify()

    def _sched_add_nrt(self, beats, task):
        # if isinstance(task, pst.PauseStream):
        #     task._next_beat = beats
        ClockTask(beats, self, task, _libsc3.main._clock_scheduler)

    def _calc_sched_beats(self, delta):
        seconds = _libsc3.main.current_tt.seconds
        beats = self.secs2beats(seconds)
        return beats + delta

    def sched(self, delta, item):
        if not self.running():
            raise ClockNotRunning(self)
        if not hasattr(item, '__awake__'):
            item = fn.Function(item)
        item._clock = self
        if self.mode == _libsc3.main.NRT_MODE:
            beats = self._calc_sched_beats(delta)
            if beats == float('inf'):
                return
            self._sched_add_nrt(beats, item)
        else:
            with self._sched_cond:
                beats = self._calc_sched_beats(delta)
                if beats == float('inf'):
                    return
                self._sched_add(beats, item)

    def sched_abs(self, beat, item):
        if not self.running():
            raise ClockNotRunning(self)
        if not hasattr(item, '__awake__'):
            item = fn.Function(item)
        item._clock = self
        if beat == float('inf'):
            return
        if self.mode == _libsc3.main.NRT_MODE:
            self._sched_add_nrt(beat, item)
        else:
            with self._sched_cond:
                self._sched_add(beat, item)

    def clear(self):
        if self.mode == _libsc3.main.NRT_MODE:
            return
        if self.running():  # and self._run_sched:  # NOTE: Was needed?
            with self._sched_cond:
                while not self._task_queue.empty():
                    self._task_queue.pop()
                self._sched_cond.notify()  # NOTE: is notify_one in C++.

    @property
    def beats_per_bar(self):
        return self._beats_per_bar

    @beats_per_bar.setter
    def beats_per_bar(self, value):
        if _libsc3.main.current_tt is not self:
            raise ClockError(
                'TempoClock should only change beats_per_bar'
                'within the scheduling thread')
        # setMeterAtBeat
        beats = self.beats
        self._base_bar = bi.round(
            (beats - self._base_bar_beat) *
            self._bars_per_beat + self._base_bar, 1)
        self._base_bar_beat = beats
        self._beats_per_bar = value
        self._bars_per_beat = 1 / value
        mdl.NotificationCenter.notify(self, 'meter')

    @property
    def base_bar_beat(self):
        return self._base_bar_beat

    @property
    def base_bar(self):
        return self._base_bar

    def beats2secs(self, beats):
        if not self.running():
            raise ClockNotRunning(self)
        return (beats - self._base_beats) * self._beat_dur + self._base_seconds

    def secs2beats(self, seconds):
        if not self.running():
            raise ClockNotRunning(self)
        return (seconds - self._base_seconds) * self._tempo + self._base_beats

    def dump(self):
        if self.running():
            print(
                f'{self.__repr__()}\n'
                f'    tempo: {self.tempo}\n'
                f'    beats: {self.beats}\n'
                f'    seconds: {self.seconds}\n'
                f'    _beat_dur: {self._beat_dur}\n'
                f'    _base_seconds: {self._base_seconds}\n'
                f'    _base_beats: {self._base_beats}')
        else:
            raise ClockNotRunning(self)

    def next_time_on_grid(self, quant=1, phase=0):
        if quant == 0:
            return self.beats + phase
        if quant < 0:
            quant = self.beats_per_bar * -quant
        if phase < 0:
            phase = bi.mod(phase, quant)
        return bi.roundup(
            self.beats - self._base_bar_beat - bi.mod(phase, quant),
            quant
        ) + self._base_bar_beat + phase

    def time_to_next_beat(self, quant=1):
        """Logical time to next beat."""
        return Quant.as_quant(quant).next_time_on_grid(self) - self.beats

    def beats2bars(self, beats):
        return (beats - self._base_bar_beat) * self._bars_per_beat\
               + self._base_bar

    def bars2beats(self, bars):
        return (bars - self._base_bar) * self._beats_per_bar\
               + self._base_bar_beat

    def bar(self):
        """Return the current bar."""
        return bi.floor(self.beats2bars(self.beats))

    def next_bar(self, beat=None):
        """Given a number of beats, determine number beats
        at the next bar line.
        """
        if beat is None:
            beat = self.beats
        return self.bars2beats(bi.ceil(self.beats2bars(beat)))

    def beat_in_bar(self):
        """Return the beat of the bar, range is 0 to < beats_per_bar."""
        return self.beats - self.bars2beats(self.bar())

    def running(self):
        if self.mode == _libsc3.main.NRT_MODE:
            return True
        else:
            return self._thread is not None and self._thread.is_alive()


def defer(func, delta=None, clock=None):
    '''
    Convenience function to defer lambda functions on a clock without
    creating a `Routine` or `sched` call. Default value for `delta` is 0.0,
    default `clock` is `AppClock`. Argument `func` can be any callable.
    '''
    if callable(func):
        def df():
            func()  # Wrapped because lambda always return its expression value.
    else:
        raise TypeError('func is not callable')
    clock = clock or AppClock
    clock.sched(delta or 0, df)



import re


### Option 1 ###

_rewrite_symbols = {
    # Valid OSC address symbols with re meaning BEFORE special symbols rewrite.
    '(': '\(',
    ')': '\)',
    '^': '\^',
    '.': '\.',
    '$': '\$',
    '+': '\+',
    '|': '\|',
    '\\': '\\\\',

    # OSC special symbols (are invalid or special OSC Address symbols).
    '{': '(?:',
    ',': '|',
    '}': ')',
    '*': '.*',  # BUG: # lo ignora el * si luego viene ?, [, { o literal, por ejemplo: '/*bc'.matchOSCAddressPattern('/abc') es false y re.match('.bc', 'abc') devuelve match. Está en el párrafo anterior al cuadro en la especificación, dice que cada caracter de pattern debe coincidir con el próximo substring de address Y que todo caracter en address debe ser emparejado con algo de pattern.
    # '[': '[',  # Same.
    # '-': '-',  # Same behaviour inside/outside brackets.
    '[!': '[^',
    # ']': ']',  # Same.
    '-]': ']', # Discard '-' before closing bracket.
    '?': '.'
}


_rewrite_pattern = re.compile(
    '(' + '|'.join(re.escape(x) for x in _rewrite_symbols.keys()) + ')')


def _rewrite_func(match):
    return _rewrite_symbols[match.group(0)]


def osc_rematch_pattern(pattern, address):
    pattern = re.sub(_rewrite_pattern, _rewrite_func, pattern)
    return re.match(pattern, address) is not None


### Option 2 ###

def osc_match_pattern(pattern, address):
    # Based on liblo lo_pattern_match in PyrSymbolPrim.cpp.
    # Iterators and sets are used because it was easier not faster.
    # Should be faster then osc_rematch_pattern with cython and indexes,
    # may have some pattern definition issues from original.
    pattern = iter(pattern)
    address = iter(address)
    p = a = None

    negate = False
    prev_p = None
    char_set = None

    p_string = None
    a_string = None
    string_set = None

    try:
        while True:
            p = next(pattern, None)

            while p == '*':
                while p == '*':  # solo para quitar los '*' consecutivos.
                    p = next(pattern, None)

                if p is None:  # '*' es el último caracter, todo lo que quede en a vale.
                    return True

                if p != '?' and p != '[' and p != '{':  # '*?', '*[abc]', '*{a|b}' leaves '*' without effect.
                    a = next(address, None)
                    while a is not None and a != p:
                        a = next(address, None)

                    p = next(pattern, None)

            if p == '?':
                a = next(address, None)
                if a is None:
                    return False
            elif p == '[':
                char_set = set()
                p = next(pattern)  # StopIteration, pattern mal formado.

                if p == '!':
                    negate = True
                    p = next(pattern)  # StopIteration, pattern mal formado.
                else:
                    negate = False

                while p != ']':
                    prev_p = p  # necesita mirar antes
                    p = next(pattern)  # StopIteration, pattern mal formado.
                    if p == '-':
                        # el guión se puede descartar.
                        p = next(pattern)  # StopIteration, pattern mal formado.
                        if p != ']':
                            # es rango
                            char_set.update(
                                chr(x) for x in range(ord(prev_p), ord(p) + 1))
                        else:
                            char_set.add(prev_p)  # es fin de corchete, el guión de más no importa.
                    else:
                        char_set.add(prev_p)
                        char_set.add(p)

                a = next(address, None)
                if a is None:
                    return False

                if negate and a in char_set:
                    return False
                elif not negate and a not in char_set:
                    return False
            elif p == '{':
                string_set = set()
                p_string = ''
                a_string = ''

                while p != '}':
                    p = next(pattern)  # StopIteration, pattern mal formado.
                    if p == ',':
                        string_set.add(p_string)
                        p_string = ''
                    elif p == '}':
                        string_set.add(p_string)
                        break
                    else:
                        p_string += p

                a = next(address, None)
                while a != '/' and a is not None:
                    a_string += a
                    if a_string in string_set:
                        break
                    a = next(address, None)

                if a == '/' or a is None:
                    return False
            else:
                a = next(address, None)
                if p != a:
                    return False

                if p is None and a is None:
                    return True
    except StopIteration:
        # pattern mal formado en algunos casos
        return False
